#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import dataclasses
from abc import abstractmethod
from typing import Any, Dict, Iterable, List, Optional, Union
from uuid import UUID

from argilla_server.enums import FieldType, MetadataPropertyType, RecordSortField, ResponseStatusFilter, SimilarityOrder
from argilla_server.models import (
    Dataset,
    Field,
    MetadataProperty,
    Question,
    QuestionType,
    Record,
    Response,
    ResponseStatus,
    Suggestion,
    Vector,
    VectorSettings,
)
from argilla_server.search_engine.base import (
    AndFilter,
    Filter,
    FilterScope,
    FloatMetadataFilter,
    FloatMetadataMetrics,
    IntegerMetadataFilter,
    IntegerMetadataMetrics,
    MetadataFilter,
    MetadataFilterScope,
    MetadataMetrics,
    Order,
    RangeFilter,
    RecordFilterScope,
    ResponseFilterScope,
    SearchEngine,
    SearchResponseItem,
    SearchResponses,
    SortBy,
    SuggestionFilterScope,
    TermsFilter,
    TermsMetadataFilter,
    TermsMetadataMetrics,
    TextQuery,
    UserResponseStatusFilter,
)

ALL_RESPONSES_STATUSES_FIELD = "all_responses_statuses"


def es_index_name_for_dataset(dataset: Dataset):
    return f"rg.{dataset.id}"


def es_terms_query(field_name: str, values: List[str]) -> dict:
    return {"terms": {field_name: values}}


def es_range_query(field_name: str, gte: Optional[float] = None, lte: Optional[float] = None) -> dict:
    query = {}
    if gte is not None:
        query["gte"] = gte
    if lte is not None:
        query["lte"] = lte
    return {"range": {field_name: query}}


def es_bool_query(
    *,
    must_not: Optional[List[dict]] = None,
    should: Optional[List[dict]] = None,
    minimum_should_match: Optional[Union[int, str]] = None,
) -> Dict[str, Any]:
    bool_query = {}

    if should:
        bool_query["should"] = should
    if must_not:
        bool_query["must_not"] = must_not

    if not bool_query:
        raise ValueError("Cannot build a boolean query without any clause")

    if minimum_should_match:
        bool_query["minimum_should_match"] = minimum_should_match

    return {"bool": bool_query}


def es_ids_query(ids: List[str]) -> dict:
    return {"ids": {"values": ids}}


def es_field_for_response_value(user: str, question: str) -> str:
    return f"responses.{user}.values.{question}"


def es_field_for_suggestion_property(question: str, property: str) -> str:
    return f"suggestions.{question}.{property}"


def es_field_for_vector_settings(vector_settings: VectorSettings) -> str:
    return f"vectors.{vector_settings.id}"


def es_field_for_record_property(property: str) -> str:
    return property


def es_field_for_metadata_property(metadata_property: Union[str, MetadataProperty]) -> str:
    if isinstance(metadata_property, MetadataProperty):
        property_name = metadata_property.name
    else:
        property_name = metadata_property

    return f"metadata.{property_name}"


def es_field_for_record_field(field_name: str) -> str:
    return f"fields.{field_name}"


def es_mapping_for_field(field: Field) -> dict:
    field_type = field.settings["type"]

    if field_type == FieldType.text:
        return {es_field_for_record_field(field.name): {"type": "text"}}
    else:
        raise Exception(f"Index configuration for field of type {field_type} cannot be generated")


def es_mapping_for_metadata_property(metadata_property: MetadataProperty) -> dict:
    property_type = metadata_property.type

    if property_type == MetadataPropertyType.terms:
        return {es_field_for_metadata_property(metadata_property): {"type": "keyword"}}
    elif property_type == MetadataPropertyType.integer:
        return {es_field_for_metadata_property(metadata_property): {"type": "long"}}
    elif property_type == MetadataPropertyType.float:
        return {es_field_for_metadata_property(metadata_property): {"type": "float"}}
    else:
        raise Exception(f"Index configuration for metadata property of type {property_type} cannot be generated")


def es_mapping_for_question(question: Question) -> dict:
    question_type = question.type

    if question_type == QuestionType.rating:
        # See https://www.elastic.co/guide/en/elasticsearch/reference/current/number.html
        return {"type": "integer"}
    elif question_type == QuestionType.text:
        # TODO: Review mapping for label selection. Could make sense to use `keyword` mapping instead.
        #  See https://www.elastic.co/guide/en/elasticsearch/reference/current/keyword.html
        #  See https://www.elastic.co/guide/en/elasticsearch/reference/current/text.html
        return {"type": "text", "index": False}
    elif question_type in [QuestionType.label_selection, QuestionType.multi_label_selection]:
        return {"type": "keyword"}
    elif question_type == QuestionType.ranking:
        return {"type": "nested", "properties": {"rank": {"type": "integer"}, "value": {"type": "keyword"}}}
    else:
        raise Exception(f"ElasticSearch mappings for Question of type {question_type} cannot be generated")


def es_mapping_for_question_suggestion(question: Question) -> dict:
    return {
        f"suggestions.{question.name}": {
            "type": "object",
            "properties": {
                "value": es_mapping_for_question(question),
                "score": {"type": "float"},
                "agent": {"type": "keyword"},
                "type": {"type": "keyword"},
            },
        }
    }


# This function will be moved once the `metadata_filters` argument is removed from search and similarity_search methods
def _unify_metadata_filters_with_filter(metadata_filters: List[MetadataFilter], filter: Optional[Filter]) -> Filter:
    filters = []
    if filter:
        filters.append(filter)

    for metadata_filter in metadata_filters:
        metadata_scope = MetadataFilterScope(metadata_property=metadata_filter.metadata_property.name)
        if isinstance(metadata_filter, TermsMetadataFilter):
            new_filter = TermsFilter(scope=metadata_scope, values=metadata_filter.values)
        elif isinstance(metadata_filter, (IntegerMetadataFilter, FloatMetadataFilter)):
            new_filter = RangeFilter(scope=metadata_scope, ge=metadata_filter.ge, le=metadata_filter.le)
        else:
            raise ValueError(f"Cannot process request for metadata filter {metadata_filter}")
        filters.append(new_filter)

    return AndFilter(filters=filters)


# This function will be moved once the response status filter is removed from search and similarity_search methods
def _unify_user_response_status_filter_with_filter(
    user_response_status_filter: UserResponseStatusFilter, filter: Optional[Filter]
) -> Filter:
    scope = ResponseFilterScope(user=user_response_status_filter.user, property="status")
    response_filter = TermsFilter(scope=scope, values=[status.value for status in user_response_status_filter.statuses])

    if filter:
        return AndFilter(filters=[filter, response_filter])
    else:
        return response_filter


# This function will be moved once the `sort_by` argument is removed from search and similarity_search methods
def _unify_sort_by_with_order(sort_by: List[SortBy], order: List[Order]) -> List[Order]:
    if order:
        return order

    new_order = []
    for sort in sort_by:
        if isinstance(sort.field, MetadataProperty):
            scope = MetadataFilterScope(metadata_property=sort.field.name)
        else:
            scope = RecordFilterScope(property=sort.field)

        new_order.append(Order(scope=scope, order=sort.order))

    return new_order


@dataclasses.dataclass
class BaseElasticAndOpenSearchEngine(SearchEngine):
    """
    Since both ElasticSearch and OpenSearch engines implementations share a lot of code,
    this class create an abstraction for the commons part of the code, letting each child
    resolve their own implementation details.

    All method for SearchEngine interface are implemented here. This abstract class defines
    some abstract method mostly for:
        1. Requesting data from the engine, since the client signatures may differ
        2. Prepare mappings for vector-related configuration (once is included)
        3. Searching records based on similarity search

    The rest of the code will be shared by both implementation
    """

    number_of_shards: int
    number_of_replicas: int

    # See https://www.elastic.co/guide/en/elasticsearch/reference/current/search-settings.html#search-settings-max-buckets
    max_terms_size: int = 2 ^ 14
    # See https://www.elastic.co/guide/en/elasticsearch/reference/5.1/index-modules.html#dynamic-index-settings
    max_result_window: int = 500000

    async def create_index(self, dataset: Dataset):
        settings = self._configure_index_settings()
        mappings = self._configure_index_mappings(dataset)

        index_name = es_index_name_for_dataset(dataset)
        await self._create_index_request(index_name, mappings, settings)

    async def configure_metadata_property(self, dataset: Dataset, metadata_property: MetadataProperty):
        mapping = es_mapping_for_metadata_property(metadata_property)
        index_name = await self._get_dataset_index(dataset)

        await self.put_index_mapping_request(index_name, mapping)

    async def delete_index(self, dataset: Dataset):
        index_name = es_index_name_for_dataset(dataset)

        await self._delete_index_request(index_name)

    async def index_records(self, dataset: Dataset, records: Iterable[Record]):
        index_name = await self._get_dataset_index(dataset)

        bulk_actions = [
            {
                # If document exist, we update source with latest version
                "_op_type": "index",  # TODO: Review and maybe change to partial update
                "_id": record.id,
                "_index": index_name,
                **self._map_record_to_es_document(record),
            }
            for record in records
        ]

        await self._bulk_op_request(bulk_actions)

    async def delete_records(self, dataset: Dataset, records: Iterable[Record]):
        index_name = await self._get_dataset_index(dataset)

        bulk_actions = [{"_op_type": "delete", "_id": record.id, "_index": index_name} for record in records]

        await self._bulk_op_request(bulk_actions)

    async def update_record_response(self, response: Response):
        record = response.record
        index_name = await self._get_dataset_index(record.dataset)

        es_responses = self._map_record_responses_to_es([response])

        await self._update_document_request(index_name, id=record.id, body={"doc": {"responses": es_responses}})

    async def delete_record_response(self, response: Response):
        record = response.record
        index_name = await self._get_dataset_index(record.dataset)

        await self._update_document_request(
            index_name, id=record.id, body={"script": f'ctx._source["responses"].remove("{response.user.username}")'}
        )

    async def update_record_suggestion(self, suggestion: Suggestion):
        index_name = await self._get_dataset_index(suggestion.record.dataset)

        es_suggestions = self._map_record_suggestions_to_es([suggestion])

        await self._update_document_request(
            index_name,
            id=suggestion.record_id,
            body={"doc": {"suggestions": es_suggestions}},
        )

    async def delete_record_suggestion(self, suggestion: Suggestion):
        index_name = await self._get_dataset_index(suggestion.record.dataset)

        await self._update_document_request(
            index_name,
            id=suggestion.record_id,
            body={"script": f'ctx._source["suggestions"].remove("{suggestion.question.name}")'},
        )

    async def set_records_vectors(self, dataset: Dataset, vectors: Iterable[Vector]):
        index_name = await self._get_dataset_index(dataset)

        bulk_actions = [
            {
                "_op_type": "update",
                "_id": vector.record_id,
                "_index": index_name,
                "doc": {es_field_for_vector_settings(vector.vector_settings): vector.value},
            }
            for vector in vectors
        ]

        await self._bulk_op_request(bulk_actions)

    async def similarity_search(
        self,
        dataset: Dataset,
        vector_settings: VectorSettings,
        value: Optional[List[float]] = None,
        record: Optional[Record] = None,
        query: Optional[Union[TextQuery, str]] = None,
        filter: Optional[Filter] = None,
        # TODO: remove them and keep filter
        user_response_status_filter: Optional[UserResponseStatusFilter] = None,
        metadata_filters: Optional[List[MetadataFilter]] = None,
        # END TODO
        max_results: int = 100,
        order: SimilarityOrder = SimilarityOrder.most_similar,
        threshold: Optional[float] = None,
    ) -> SearchResponses:
        # TODO: This block will be moved (maybe to contexts/search.py), and only filter and order arguments will be kept
        if metadata_filters:
            filter = _unify_metadata_filters_with_filter(metadata_filters, filter)
        if user_response_status_filter and user_response_status_filter.statuses:
            filter = _unify_user_response_status_filter_with_filter(user_response_status_filter, filter)
        # END TODO

        if bool(value) == bool(record):
            raise ValueError("Must provide either vector value or record to compute the similarity search")

        vector_value = value
        record_id = None

        if not vector_value:
            record_id = record.id
            vector_value = record.vector_value_by_vector_settings(vector_settings)

        if not vector_value:
            raise ValueError("Cannot find a vector value to apply with provided info")

        if order == SimilarityOrder.least_similar:
            vector_value = self._inverse_vector(vector_value)

        query_filters = []
        if filter:
            # Wrapping filter in a list to use easily on each engine implementation
            query_filters = [self.build_elasticsearch_filter(filter)]

        index = await self._get_dataset_index(dataset)
        response = await self._request_similarity_search(
            index=index,
            vector_settings=vector_settings,
            value=vector_value,
            k=max_results,
            excluded_id=record_id,
            query_filters=query_filters,
        )

        return await self._process_search_response(response, threshold)

    def build_elasticsearch_filter(self, filter: Filter) -> Dict[str, Any]:
        def is_response_status_scope(scope: FilterScope) -> bool:
            if not isinstance(scope, ResponseFilterScope):
                return False

            if not scope.property == "status":
                return False

            if scope.question is not None:
                return False

            return True

        if isinstance(filter, AndFilter):
            filters = [self.build_elasticsearch_filter(f) for f in filter.filters]
            return es_bool_query(should=filters, minimum_should_match=len(filters))

        # This is a special case for response status filter, since it's compound by multiple filters
        if is_response_status_scope(filter.scope):
            status_filter = UserResponseStatusFilter(
                user=filter.scope.user, statuses=[ResponseStatusFilter(v) for v in filter.values]
            )
            return self._build_response_status_filter(status_filter)

        es_field = self._scope_to_elasticsearch_field(filter.scope)

        if isinstance(filter, TermsFilter):
            return es_terms_query(es_field, values=filter.values)
        elif isinstance(filter, RangeFilter):
            return es_range_query(es_field, gte=filter.ge, lte=filter.le)
        else:
            raise ValueError(f"Cannot process request for filter {filter}")

    def build_elasticsearch_sort(self, sort: List[Order]) -> str:
        sort_config = []

        for order in sort:
            sort_field_name = self._scope_to_elasticsearch_field(order.scope)
            sort_config.append(f"{sort_field_name}:{order.order}")

        return ",".join(sort_config)

    @staticmethod
    def _scope_to_elasticsearch_field(scope: FilterScope) -> str:
        if isinstance(scope, MetadataFilterScope):
            return es_field_for_metadata_property(scope.metadata_property)
        elif isinstance(scope, SuggestionFilterScope):
            return es_field_for_suggestion_property(question=scope.question, property=scope.property)
        elif isinstance(scope, ResponseFilterScope):
            return es_field_for_response_value(scope.user.username, question=scope.question)
        elif isinstance(scope, RecordFilterScope):
            return es_field_for_record_property(scope.property)
        raise ValueError(f"Cannot process request for search scope {scope}")

    @staticmethod
    def _build_response_status_filter(status_filter: UserResponseStatusFilter) -> Dict[str, Any]:
        if status_filter.user is None:
            response_field = ALL_RESPONSES_STATUSES_FIELD
        else:
            response_field = f"responses.{status_filter.user.username}.status"

        filters = []
        if status_filter.has_pending_status:
            # See https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-exists-query.html
            filters.append({"bool": {"must_not": {"exists": {"field": response_field}}}})

        if status_filter.response_statuses:
            # See https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-terms-query.html
            filters.append(es_terms_query(response_field, values=status_filter.response_statuses))

        return {"bool": {"should": filters, "minimum_should_match": 1}}

    def _inverse_vector(self, vector_value: List[float]) -> List[float]:
        return [vector_value[i] * -1 for i in range(0, len(vector_value))]

    def _map_record_to_es_document(self, record: Record) -> Dict[str, Any]:
        document = {
            "id": str(record.id),
            "fields": record.fields,
            "inserted_at": record.inserted_at,
            "updated_at": record.updated_at,
        }

        if record.metadata_:
            document["metadata"] = self._map_record_metadata_to_es(record.metadata_, record.dataset.metadata_properties)
        if record.responses:
            document["responses"] = self._map_record_responses_to_es(record.responses)
        if record.suggestions:
            document["suggestions"] = self._map_record_suggestions_to_es(record.suggestions)
        if record.vectors:
            document["vectors"] = self._map_record_vectors_to_es(record.vectors)

        return document

    @staticmethod
    def _map_record_responses_to_es(responses: List[Response]) -> Dict[str, Any]:
        return {
            response.user.username: {
                "values": {k: v["value"] for k, v in response.values.items()} if response.values else None,
                "status": response.status,
            }
            for response in responses
        }

    @staticmethod
    def _map_record_suggestions_to_es(suggestions: List[Suggestion]) -> Dict[str, str]:
        return {
            suggestion.question.name: {
                "type": suggestion.type,
                "agent": suggestion.agent,
                "score": suggestion.score,
                "value": suggestion.value,
            }
            for suggestion in suggestions
        }

    @staticmethod
    def _map_record_vectors_to_es(vectors: List[Vector]) -> Dict[str, List[float]]:
        return {str(vector.vector_settings.id): vector.value for vector in vectors}

    @staticmethod
    def _map_record_metadata_to_es(
        metadata: Dict[str, Any], metadata_properties: List[MetadataProperty]
    ) -> Dict[str, Any]:
        search_engine_metadata = {}

        for metadata_property in metadata_properties:
            value = metadata.get(metadata_property.name)
            if value is not None:
                search_engine_metadata[str(metadata_property.name)] = value

        return search_engine_metadata

    async def configure_index_vectors(self, vector_settings: VectorSettings) -> None:
        index = await self._get_dataset_index(vector_settings.dataset)

        mappings = self._mapping_for_vector_settings(vector_settings)
        await self.put_index_mapping_request(index, mappings)

    async def search(
        self,
        dataset: Dataset,
        query: Optional[Union[TextQuery, str]] = None,
        filter: Optional[Filter] = None,
        sort: Optional[List[Order]] = None,
        # TODO: Remove these arguments
        user_response_status_filter: Optional[UserResponseStatusFilter] = None,
        metadata_filters: Optional[List[MetadataFilter]] = None,
        sort_by: Optional[List[SortBy]] = None,
        # END TODO
        offset: int = 0,
        limit: int = 100,
    ) -> SearchResponses:
        # See https://www.elastic.co/guide/en/elasticsearch/reference/current/search-search.html

        # TODO: This block will be moved (maybe to contexts/search.py), and only filter and order arguments will be kept
        if metadata_filters:
            filter = _unify_metadata_filters_with_filter(metadata_filters, filter)
        if user_response_status_filter and user_response_status_filter.statuses:
            filter = _unify_user_response_status_filter_with_filter(user_response_status_filter, filter)

        if sort_by:
            sort = _unify_sort_by_with_order(sort_by, sort)
        # END TODO

        text_query = self._build_text_query(dataset, text=query)
        bool_query: Dict[str, Any] = {"must": [text_query]}

        if filter:
            bool_query["filter"] = self.build_elasticsearch_filter(filter)

        es_query = {"bool": bool_query}
        index = await self._get_dataset_index(dataset)

        es_sort = self.build_elasticsearch_sort(sort) if sort else None
        response = await self._index_search_request(index, query=es_query, size=limit, from_=offset, sort=es_sort)

        return await self._process_search_response(response)

    async def compute_metrics_for(self, metadata_property: MetadataProperty) -> MetadataMetrics:
        index_name = await self._get_dataset_index(metadata_property.dataset)

        if metadata_property.type == MetadataPropertyType.terms:
            return await self._metrics_for_terms_property(index_name, metadata_property)

        if metadata_property.type in [MetadataPropertyType.float, MetadataPropertyType.integer]:
            return await self._metrics_for_numeric_property(index_name, metadata_property)

    async def _metrics_for_numeric_property(
        self, index_name: str, metadata_property: MetadataProperty, query: Optional[dict] = None
    ) -> Union[IntegerMetadataMetrics, FloatMetadataMetrics]:
        field_name = es_field_for_metadata_property(metadata_property)
        query = query or {"match_all": {}}

        stats = await self.__stats_aggregation(index_name, field_name, query)

        metrics_class = (
            IntegerMetadataMetrics if metadata_property.type == MetadataPropertyType.integer else FloatMetadataMetrics
        )

        return metrics_class(min=stats["min"], max=stats["max"])

    async def _metrics_for_terms_property(
        self, index_name: str, metadata_property: MetadataProperty, query: Optional[dict] = None
    ) -> TermsMetadataMetrics:
        field_name = es_field_for_metadata_property(metadata_property)
        query = query or {"match_all": {}}

        total_terms = await self.__value_count_aggregation(index_name, field_name=field_name, query=query)
        if total_terms == 0:
            return TermsMetadataMetrics(total=total_terms)

        terms_buckets = await self.__terms_aggregation(index_name, field_name=field_name, query=query, size=total_terms)
        terms_values = [
            TermsMetadataMetrics.TermCount(term=bucket["key"], count=bucket["doc_count"]) for bucket in terms_buckets
        ]
        return TermsMetadataMetrics(total=total_terms, values=terms_values)

    def _configure_index_mappings(self, dataset: Dataset) -> dict:
        return {
            # See https://www.elastic.co/guide/en/elasticsearch/reference/current/dynamic.html#dynamic-parameters
            "dynamic": "strict",
            "dynamic_templates": self._dynamic_templates_for_question_responses(dataset.questions),
            "properties": {
                # See https://www.elastic.co/guide/en/elasticsearch/reference/current/explicit-mapping.html
                "id": {"type": "keyword"},
                RecordSortField.inserted_at.value: {"type": "date_nanos"},
                RecordSortField.updated_at.value: {"type": "date_nanos"},
                "responses": {"dynamic": True, "type": "object"},
                ALL_RESPONSES_STATUSES_FIELD: {"type": "keyword"},  # To add all users responses
                **self._mapping_for_fields(dataset.fields),
                **self._mapping_for_suggestions(dataset.questions),
                **self._mapping_for_metadata_properties(dataset.metadata_properties),
                **self._mapping_for_vectors_settings(dataset.vectors_settings),
            },
        }

    async def _process_search_response(
        self, response: dict, score_threshold: Optional[float] = None
    ) -> SearchResponses:
        hits = response["hits"]["hits"]

        if score_threshold is not None:
            hits = filter(lambda hit: hit["_score"] >= score_threshold, hits)

        items = [SearchResponseItem(record_id=UUID(hit["_id"]), score=hit["_score"]) for hit in hits]
        total = response["hits"]["total"]["value"]

        return SearchResponses(items=items, total=total)

    @staticmethod
    def _build_text_query(dataset: Dataset, text: Optional[Union[TextQuery, str]] = None) -> dict:
        if text is None:
            return {"match_all": {}}

        if isinstance(text, str):
            text = TextQuery(q=text)

        if not text.field:
            # See https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-multi-match-query.html
            field_names = [
                es_field_for_record_field(field.name)
                for field in dataset.fields
                if field.settings.get("type") == FieldType.text
            ]
            return {"multi_match": {"query": text.q, "type": "cross_fields", "fields": field_names, "operator": "and"}}
        else:
            # See https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-match-query.html
            return {"match": {es_field_for_record_field(text.field): {"query": text.q, "operator": "and"}}}

    def _mapping_for_fields(self, fields: List[Field]) -> dict:
        mappings = {}
        for field in fields:
            mappings.update(es_mapping_for_field(field))

        return mappings

    def _mapping_for_metadata_properties(self, metadata_properties: List[MetadataProperty]) -> dict:
        mappings = {
            # metadata properties without mappings will be ignored
            "metadata": {"dynamic": False, "type": "object"},
        }

        for metadata_property in metadata_properties:
            mappings.update(es_mapping_for_metadata_property(metadata_property))

        return mappings

    def _mapping_for_suggestions(self, questions: List[Question]) -> dict:
        mappings = {}

        for question in questions:
            mappings.update(es_mapping_for_question_suggestion(question))

        return mappings

    def _dynamic_templates_for_question_responses(self, questions: List[Question]) -> List[dict]:
        # See https://www.elastic.co/guide/en/elasticsearch/reference/current/dynamic-templates.html
        return [
            {
                "status_responses": {
                    "path_match": "responses.*.status",
                    "mapping": {"type": "keyword", "copy_to": ALL_RESPONSES_STATUSES_FIELD},
                }
            },
            *[
                {
                    f"{question.name}_responses": {
                        "path_match": f"responses.*.values.{question.name}",
                        "mapping": es_mapping_for_question(question),
                    },
                }
                for question in questions
            ],
        ]

    async def _get_dataset_index(self, dataset: Dataset):
        index_name = es_index_name_for_dataset(dataset)

        return index_name

    def _mapping_for_vectors_settings(self, vectors_settings: List[VectorSettings]) -> dict:
        mappings = {}
        for vector in vectors_settings:
            mappings.update(self._mapping_for_vector_settings(vector))

        return mappings

    async def __terms_aggregation(self, index_name: str, field_name: str, query: dict, size: int) -> List[dict]:
        aggregation_name = "terms_agg"

        terms_agg = {aggregation_name: {"terms": {"field": field_name, "size": min(size, self.max_terms_size)}}}

        response = await self._index_search_request(index_name, query=query, aggregations=terms_agg, size=0)
        return response["aggregations"][aggregation_name]["buckets"]

    async def __value_count_aggregation(self, index_name: str, field_name: str, query: dict) -> int:
        aggregation_name = "count_values"

        value_count_agg = {aggregation_name: {"value_count": {"field": field_name}}}

        response = await self._index_search_request(index_name, query=query, aggregations=value_count_agg, size=0)
        return response["aggregations"][aggregation_name]["value"]

    async def __stats_aggregation(self, index_name: str, field_name: str, query: dict) -> dict:
        # See https://www.elastic.co/guide/en/elasticsearch/reference/current/search-aggregations-metrics-stats-aggregation.html
        aggregation_name = f"numeric_stats"

        stats_agg = {aggregation_name: {"stats": {"field": field_name}}}

        response = await self._index_search_request(index_name, query=query, aggregations=stats_agg, size=0)
        return response["aggregations"][aggregation_name]

    @abstractmethod
    def _configure_index_settings(self) -> dict:
        """Defines settings configuration for the index. Depending on which backend is used, this may differ"""
        pass

    @abstractmethod
    def _mapping_for_vector_settings(self, vector_settings: VectorSettings) -> dict:
        """Defines one mapping property configuration for a vector_setting definition"""
        pass

    @abstractmethod
    async def _request_similarity_search(
        self,
        index: str,
        vector_settings: VectorSettings,
        value: List[float],
        k: int,
        excluded_id: Optional[UUID] = None,
        query_filters: Optional[List[dict]] = None,
    ) -> dict:
        """
        Applies the similarity search request based on a vector configuration, a vector value,
        the `k` number of results to retrieve and an optional filter configuration to apply
        """
        pass

    @abstractmethod
    async def _create_index_request(self, index_name: str, mappings: dict, settings: dict) -> None:
        """Executes request for index creation"""
        pass

    @abstractmethod
    async def _delete_index_request(self, index_name: str):
        """Executes request for index deletion"""
        pass

    @abstractmethod
    async def _update_document_request(self, index_name: str, id: str, body: dict):
        """Executes request for index document (partial) update"""
        pass

    @abstractmethod
    async def put_index_mapping_request(self, index: str, mappings: dict):
        """Executes request for index mapping (partial) update"""
        pass

    @abstractmethod
    async def _index_search_request(
        self,
        index: str,
        query: dict,
        size: Optional[int] = None,
        from_: Optional[int] = None,
        sort: Optional[str] = None,
        aggregations: Optional[dict] = None,
    ) -> dict:
        """Executes request for search documents on a index"""
        pass

    @abstractmethod
    async def _index_exists_request(self, index_name: str) -> bool:
        """Executes request for check if index exists"""
        pass

    @abstractmethod
    async def _bulk_op_request(self, actions: List[Dict[str, Any]]):
        """Executes request for bulk operations"""
        pass

    @abstractmethod
    async def _refresh_index_request(self, index_name: str):
        pass



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import TYPE_CHECKING

import pytest
from argilla_server.apis.v0.models.text2text import Text2TextBulkRequest, Text2TextRecord, Text2TextRecordInputs
from argilla_server.apis.v0.models.text_classification import TextClassificationBulkRequest, TextClassificationRecord
from argilla_server.apis.v0.models.token_classification import TokenClassificationBulkRequest, TokenClassificationRecord
from argilla_server.commons.models import TaskType
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.models import User, UserRole

from tests.factories import UserFactory, WorkspaceFactory

if TYPE_CHECKING:
    from httpx import AsyncClient


async def create_dataset(async_client: "AsyncClient", workspace_name: str, task: TaskType) -> str:
    dataset_name = "test_dataset"
    params = {"workspace": workspace_name}

    response = await async_client.delete(f"/api/datasets/{dataset_name}", params=params)
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets",
        json={"name": dataset_name, "task": task, "workspace": workspace_name},
    )
    assert response.status_code == 200

    def text_classification_request():
        return TextClassificationBulkRequest(
            records=[
                TextClassificationRecord(
                    **{
                        "id": 0,
                        "inputs": {"data": "my data"},
                        "prediction": {
                            "agent": "test",
                            "labels": [
                                {"class": "Test", "score": 0.3},
                                {"class": "Mocking", "score": 0.7},
                            ],
                        },
                    }
                )
            ],
        )

    def token_classification_request():
        return TokenClassificationBulkRequest(
            records=[
                TokenClassificationRecord.parse_obj(data)
                for data in [
                    {
                        "id": 0,
                        "tokens": "This is a text".split(" "),
                        "raw_text": "This is a text",
                        "metadata": {"field_one": "value one", "field_two": "value 2"},
                        "prediction": {
                            "agent": "test",
                            "entities": [
                                {"start": 0, "end": 4, "label": "PERSON"},
                                {"start": 5, "end": 7, "label": "ORG"},
                            ],
                        },
                    },
                    {
                        "id": 1,
                        "tokens": "This is a text".split(" "),
                        "raw_text": "This is a text",
                        "metadata": {"field_one": "value one", "field_two": "value 2"},
                        "prediction": {"agent": "test", "entities": [{"start": 0, "end": 4, "label": "PERSON"}]},
                    },
                    {
                        "id": 2,
                        "tokens": "This is a text".split(" "),
                        "raw_text": "This is a text",
                        "metadata": {"field_one": "value one", "field_two": "value 2"},
                        "annotation": {"agent": "test", "entities": [{"start": 0, "end": 4, "label": "ORG"}]},
                    },
                ]
            ],
        )

    def text2text_request():
        return Text2TextBulkRequest(
            records=[
                Text2TextRecordInputs.parse_obj(
                    {
                        "id": 0,
                        "text": "This is a text data",
                        "predictions": {"test": {"sentences": [{"text": "This is a test data", "score": 0.6}]}},
                    }
                ),
                Text2TextRecordInputs.parse_obj(
                    {
                        "id": 1,
                        "text": "Another data",
                        "annotations": {
                            "annotator-1": {"sentences": [{"text": "THis is a test data"}]},
                            "annotator-2": {"sentences": [{"text": "This IS the test datay"}]},
                        },
                    }
                ),
            ]
        )

    request_builder = {
        TaskType.text2text: text2text_request,
        TaskType.token_classification: token_classification_request,
        TaskType.text_classification: text_classification_request,
    }[task]

    response = await async_client.post(
        f"/api/datasets/{dataset_name}/{task}:bulk",
        json=request_builder().dict(by_alias=True, exclude_none=True),
        params=params,
    )
    assert response.status_code == 200

    return dataset_name


@pytest.mark.parametrize("role", [UserRole.owner, UserRole.admin])
@pytest.mark.parametrize(
    ("task", "expected_record_class"),
    [
        (TaskType.token_classification, TokenClassificationRecord),
        (TaskType.text_classification, TextClassificationRecord),
        (TaskType.text2text, Text2TextRecord),
    ],
)
@pytest.mark.asyncio
async def test_update_record_ok(async_client: "AsyncClient", owner: User, task, expected_record_class, role: UserRole):
    async_client.headers.update({API_KEY_HEADER_NAME: owner.api_key})

    expected_id = 0
    workspace = await WorkspaceFactory.create()
    dataset = await create_dataset(async_client, workspace_name=workspace.name, task=task)
    user = await UserFactory.create(role=role, workspaces=[workspace])

    text = "A new value for record 0"
    response = await async_client.patch(
        f"/api/datasets/{dataset}/records/{expected_id}",
        json={"metadata": {"new_value": text}},
        headers={API_KEY_HEADER_NAME: user.api_key},
        params={"workspace": workspace.name},
    )
    assert response.status_code == 200, response.json()
    updated_record = expected_record_class.parse_obj(response.json())
    assert updated_record.metadata["new_value"] == text

    response = await async_client.post(
        f"/api/datasets/{dataset}/{task}:search",
        json={"query": {"ids": [expected_id]}},
        headers={API_KEY_HEADER_NAME: user.api_key},
        params={"workspace": workspace.name},
    )
    assert response.status_code == 200
    data = response.json()
    assert data["total"] == 1
    record = data["records"][0]

    updated_record.metrics = {}
    assert updated_record == expected_record_class.parse_obj(record)


@pytest.mark.parametrize("role", [UserRole.owner, UserRole.admin])
@pytest.mark.parametrize(
    "task",
    [
        (TaskType.token_classification),
        (TaskType.text_classification),
        (TaskType.text2text),
    ],
)
@pytest.mark.asyncio
async def test_update_with_not_found(async_client: "AsyncClient", task: TaskType, role: UserRole):
    workspace = await WorkspaceFactory.create()
    user = await UserFactory.create(role=role, workspaces=[workspace])

    async_client.headers.update({API_KEY_HEADER_NAME: user.api_key})

    dataset = await create_dataset(async_client, workspace_name=workspace.name, task=task)

    response = await async_client.patch(
        f"/api/datasets/{dataset}/records/not-found",
        json={"status": "Validated"},
        headers={API_KEY_HEADER_NAME: user.api_key},
        params={"workspace": workspace.name},
    )
    assert response.status_code == 404
    assert response.json() == {
        "detail": {
            "code": "argilla.api.errors::RecordNotFound",
            "params": {
                "dataset": f"{workspace.name}.test_dataset",
                "id": "not-found",
                "name": f"{workspace.name}.test_dataset.not-found",
                "type": "Record",
            },
        }
    }


@pytest.mark.asyncio
async def test_with_wrong_values(async_client: "AsyncClient", owner: User):
    async_client.headers.update({API_KEY_HEADER_NAME: owner.api_key})
    workspace = await WorkspaceFactory.create()
    dataset = await create_dataset(async_client, workspace_name=workspace.name, task=TaskType.text_classification)

    response = await async_client.patch(
        f"/api/datasets/{dataset}/records/not-found",
        json={"status": "NOStatus"},
        headers={API_KEY_HEADER_NAME: owner.api_key},
        params={"workspace": workspace.name},
    )

    assert response.status_code == 422



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Common model for task definitions
"""

from typing import Any, Dict, Generic, List, TypeVar

from argilla_server.pydantic_v1 import BaseModel, Field
from argilla_server.pydantic_v1.generics import GenericModel
from argilla_server.services.search.model import (
    ServiceQueryRange,
    ServiceSearchResultsAggregations,
    ServiceSortableField,
)
from argilla_server.services.tasks.commons import ServiceBaseAnnotation, ServiceBaseRecord, ServiceBaseRecordInputs


class SortableField(ServiceSortableField):
    pass


class BaseAnnotation(ServiceBaseAnnotation):
    pass


Annotation = TypeVar("Annotation", bound=BaseAnnotation)


class BaseRecordInputs(ServiceBaseRecordInputs[Annotation], Generic[Annotation]):
    def extended_fields(self) -> Dict[str, Any]:
        return {}


class BaseRecord(ServiceBaseRecord[Annotation], Generic[Annotation]):
    pass


class ScoreRange(ServiceQueryRange):
    pass


_Record = TypeVar("_Record", bound=BaseRecord)


class BulkResponse(BaseModel):
    dataset: str
    processed: int
    failed: int = 0


class BaseSearchResults(GenericModel, Generic[_Record, ServiceSearchResultsAggregations]):
    total: int = 0
    records: List[_Record] = Field(default_factory=list)
    aggregations: ServiceSearchResultsAggregations = None



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import TYPE_CHECKING, Any, Callable, Type
from unittest.mock import call
from uuid import UUID, uuid4

import pytest
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.enums import ResponseStatus
from argilla_server.models import Dataset, Record, Response, Suggestion, User, UserRole
from argilla_server.search_engine import SearchEngine
from sqlalchemy import func, select
from sqlalchemy.orm import Session

from tests.factories import (
    DatasetFactory,
    FloatMetadataPropertyFactory,
    IntegerMetadataPropertyFactory,
    LabelSelectionQuestionFactory,
    MetadataPropertyFactory,
    MultiLabelSelectionQuestionFactory,
    QuestionFactory,
    RankingQuestionFactory,
    RatingQuestionFactory,
    RecordFactory,
    ResponseFactory,
    SuggestionFactory,
    TermsMetadataPropertyFactory,
    TextQuestionFactory,
    UserFactory,
    VectorFactory,
    VectorSettingsFactory,
    WorkspaceFactory,
)

if TYPE_CHECKING:
    from argilla_server.models import Dataset
    from httpx import AsyncClient
    from sqlalchemy.ext.asyncio import AsyncSession


async def create_text_questions(dataset: "Dataset") -> None:
    await TextQuestionFactory.create(name="input_ok", dataset=dataset, required=True)
    await TextQuestionFactory.create(name="output_ok", dataset=dataset)


async def create_rating_questions(dataset: "Dataset") -> None:
    await RatingQuestionFactory.create(name="rating_question_1", dataset=dataset, required=True)
    await RatingQuestionFactory.create(name="rating_question_2", dataset=dataset)


async def create_label_selection_questions(dataset: "Dataset") -> None:
    await LabelSelectionQuestionFactory.create(name="label_selection_question_1", dataset=dataset, required=True)
    await LabelSelectionQuestionFactory.create(name="label_selection_question_2", dataset=dataset)


async def create_multi_label_selection_questions(dataset: "Dataset") -> None:
    await MultiLabelSelectionQuestionFactory.create(
        name="multi_label_selection_question_1", dataset=dataset, required=True
    )
    await MultiLabelSelectionQuestionFactory.create(name="multi_label_selection_question_2", dataset=dataset)


async def create_ranking_question(dataset: "Dataset") -> None:
    await RankingQuestionFactory.create(name="ranking_question_1", dataset=dataset, required=True)
    await RankingQuestionFactory.create(name="ranking_question_2", dataset=dataset)


@pytest.mark.asyncio
class TestSuiteRecords:
    @pytest.mark.parametrize("role", [UserRole.owner, UserRole.admin, UserRole.annotator])
    async def test_get_record(self, async_client: "AsyncClient", role: UserRole):
        dataset = await DatasetFactory.create()
        record = await RecordFactory.create(dataset=dataset)
        user = await UserFactory.create(workspaces=[dataset.workspace], role=role)

        response = await async_client.get(f"/api/v1/records/{record.id}", headers={API_KEY_HEADER_NAME: user.api_key})

        assert response.status_code == 200
        assert response.json() == {
            "id": str(record.id),
            "fields": {"text": "This is a text", "sentiment": "neutral"},
            "metadata": None,
            "external_id": record.external_id,
            "responses": None,
            "suggestions": [],
            "vectors": None,
            "dataset_id": str(dataset.id),
            "inserted_at": record.inserted_at.isoformat(),
            "updated_at": record.updated_at.isoformat(),
        }

    async def test_get_record_without_authentication(self, async_client: "AsyncClient"):
        record = await RecordFactory.create()

        response = await async_client.get(f"/api/v1/records/{record.id}")

        assert response.status_code == 401

    @pytest.mark.parametrize("role", [UserRole.admin, UserRole.annotator])
    async def test_get_record_as_restricted_user_from_different_workspace(
        self, async_client: "AsyncClient", role: UserRole
    ):
        dataset = await DatasetFactory.create()
        record = await RecordFactory.create(dataset=dataset)
        user = await UserFactory.create(role=role)

        response = await async_client.get(f"/api/v1/records/{record.id}", headers={API_KEY_HEADER_NAME: user.api_key})

        assert response.status_code == 403

    async def test_get_record_with_nonexistent_record_id(self, async_client: "AsyncClient", owner_auth_header: dict):
        await RecordFactory.create()

        response = await async_client.get(f"/api/v1/records/{uuid4()}", headers=owner_auth_header)

        assert response.status_code == 404

    @pytest.mark.parametrize("role", [UserRole.owner, UserRole.admin])
    async def test_update_record(self, async_client: "AsyncClient", mock_search_engine: SearchEngine, role: UserRole):
        dataset = await DatasetFactory.create()
        user = await UserFactory.create(workspaces=[dataset.workspace], role=role)
        question_0 = await TextQuestionFactory.create(dataset=dataset)
        question_1 = await TextQuestionFactory.create(dataset=dataset)
        question_2 = await TextQuestionFactory.create(dataset=dataset)
        await TermsMetadataPropertyFactory.create(name="terms-metadata-property", dataset=dataset)
        await IntegerMetadataPropertyFactory.create(name="integer-metadata-property", dataset=dataset)
        await FloatMetadataPropertyFactory.create(name="float-metadata-property", dataset=dataset)
        record = await RecordFactory.create(
            dataset=dataset,
            metadata_={"terms-metadata-property": "a", "integer-metadata-property": 1, "float-metadata-property": 1.0},
        )
        await SuggestionFactory.create(question=question_0, record=record, value="suggestion 1")
        await SuggestionFactory.create(question=question_1, record=record, value="suggestion 2")
        await SuggestionFactory.create(question=question_2, record=record, value="suggestion 3")
        vector_settings_0 = await VectorSettingsFactory.create(dataset=dataset, dimensions=5)
        vector_settings_1 = await VectorSettingsFactory.create(dataset=dataset, dimensions=5)
        vector_settings_2 = await VectorSettingsFactory.create(dataset=dataset, dimensions=5)
        await VectorFactory.create(record=record, vector_settings=vector_settings_0, value=[1, 1, 1, 1, 1])
        await VectorFactory.create(record=record, vector_settings=vector_settings_1, value=[2, 2, 2, 2, 2])

        response = await async_client.patch(
            f"/api/v1/records/{record.id}",
            headers={API_KEY_HEADER_NAME: user.api_key},
            json={
                "metadata": {
                    "terms-metadata-property": "c",
                    "integer-metadata-property": 9999,
                    "float-metadata-property": 9999.0,
                    "extra-metadata": "yes",
                },
                "suggestions": [
                    {
                        "question_id": str(question_0.id),
                        "value": "suggestion updated 1",
                    },
                    {
                        "question_id": str(question_1.id),
                        "value": "suggestion updated 2",
                    },
                ],
                "vectors": {
                    vector_settings_0.name: [10, 10, 10, 10, 10],
                    vector_settings_2.name: [3, 3, 3, 3, 3],
                },
            },
        )

        assert response.status_code == 200
        assert response.json() == {
            "id": str(record.id),
            "fields": {"text": "This is a text", "sentiment": "neutral"},
            "metadata": {
                "terms-metadata-property": "c",
                "integer-metadata-property": 9999,
                "float-metadata-property": 9999.0,
                "extra-metadata": "yes",
            },
            "external_id": record.external_id,
            "responses": [],
            "suggestions": [
                {
                    "question_id": str(question_0.id),
                    "type": None,
                    "score": None,
                    "value": "suggestion updated 1",
                    "agent": None,
                    "id": str(record.suggestions[0].id),
                },
                {
                    "question_id": str(question_1.id),
                    "type": None,
                    "score": None,
                    "value": "suggestion updated 2",
                    "agent": None,
                    "id": str(record.suggestions[1].id),
                },
            ],
            "vectors": {
                vector_settings_0.name: [10.0, 10.0, 10.0, 10.0, 10.0],
                vector_settings_1.name: [2.0, 2.0, 2.0, 2.0, 2.0],
                vector_settings_2.name: [3.0, 3.0, 3.0, 3.0, 3.0],
            },
            "dataset_id": str(dataset.id),
            "inserted_at": record.inserted_at.isoformat(),
            "updated_at": record.updated_at.isoformat(),
        }
        mock_search_engine.index_records.assert_called_once_with(dataset, [record])

    async def test_update_record_with_null_metadata(
        self, async_client: "AsyncClient", mock_search_engine: SearchEngine, owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()
        await TermsMetadataPropertyFactory.create(name="terms-metadata-property", dataset=dataset)
        await IntegerMetadataPropertyFactory.create(name="integer-metadata-property", dataset=dataset)
        await FloatMetadataPropertyFactory.create(name="float-metadata-property", dataset=dataset)
        record = await RecordFactory.create(
            dataset=dataset,
            metadata_={"terms-metadata-property": "a", "integer-metadata-property": 1, "float-metadata-property": 1.0},
        )

        response = await async_client.patch(
            f"/api/v1/records/{record.id}",
            headers=owner_auth_header,
            json={"metadata": None},
        )

        assert response.status_code == 200
        assert response.json() == {
            "id": str(record.id),
            "fields": {"text": "This is a text", "sentiment": "neutral"},
            "metadata": None,
            "external_id": record.external_id,
            "responses": [],
            "suggestions": [],
            "vectors": {},
            "dataset_id": str(dataset.id),
            "inserted_at": record.inserted_at.isoformat(),
            "updated_at": record.updated_at.isoformat(),
        }
        mock_search_engine.index_records.assert_called_once_with(dataset, [record])

    async def test_update_record_with_no_metadata(
        self, async_client: "AsyncClient", mock_search_engine: SearchEngine, owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()
        record = await RecordFactory.create(dataset=dataset)

        response = await async_client.patch(
            f"/api/v1/records/{record.id}",
            headers=owner_auth_header,
            json={},
        )

        assert response.status_code == 200
        assert response.json() == {
            "id": str(record.id),
            "fields": {"text": "This is a text", "sentiment": "neutral"},
            "metadata": None,
            "external_id": record.external_id,
            "responses": None,
            "suggestions": [],
            "vectors": {},
            "dataset_id": str(dataset.id),
            "inserted_at": record.inserted_at.isoformat(),
            "updated_at": record.updated_at.isoformat(),
        }
        mock_search_engine.index_records.assert_not_called()

    async def test_update_record_with_list_terms_metadata(
        self, async_client: "AsyncClient", mock_search_engine: SearchEngine, owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()
        await TermsMetadataPropertyFactory.create(name="terms-metadata-property", dataset=dataset)
        record = await RecordFactory.create(dataset=dataset)

        response = await async_client.patch(
            f"/api/v1/records/{record.id}",
            headers=owner_auth_header,
            json={
                "metadata": {
                    "terms-metadata-property": ["a", "b", "c"],
                },
            },
        )

        assert response.status_code == 200
        assert response.json() == {
            "id": str(record.id),
            "fields": {"text": "This is a text", "sentiment": "neutral"},
            "metadata": {
                "terms-metadata-property": ["a", "b", "c"],
            },
            "external_id": record.external_id,
            "responses": [],
            "suggestions": [],
            "vectors": {},
            "dataset_id": str(dataset.id),
            "inserted_at": record.inserted_at.isoformat(),
            "updated_at": record.updated_at.isoformat(),
        }
        mock_search_engine.index_records.assert_called_once_with(dataset, [record])

    async def test_update_record_with_no_suggestions(
        self, async_client: "AsyncClient", db: "AsyncSession", mock_search_engine: SearchEngine, owner_auth_header: dict
    ):
        suggestion = await SuggestionFactory.create()
        record = suggestion.record

        response = await async_client.patch(
            f"/api/v1/records/{suggestion.record.id}",
            headers=owner_auth_header,
            json={"suggestions": []},
        )

        assert response.status_code == 200
        assert response.json() == {
            "id": str(record.id),
            "fields": {"text": "This is a text", "sentiment": "neutral"},
            "metadata": None,
            "external_id": record.external_id,
            "responses": None,
            "suggestions": [],
            "vectors": {},
            "dataset_id": str(record.dataset_id),
            "inserted_at": record.inserted_at.isoformat(),
            "updated_at": record.updated_at.isoformat(),
        }
        assert (await db.execute(select(Suggestion).where(Suggestion.id == suggestion.id))).scalar_one_or_none() is None

    @pytest.mark.parametrize(
        ["MetadataPropertyFactoryClass", "create_value", "update_value", "expected_error"],
        [
            (
                TermsMetadataPropertyFactory,
                "a",
                "z",
                "metadata is not valid: 'name' metadata property validation failed because 'z' is not an allowed term.",
            ),
            (
                IntegerMetadataPropertyFactory,
                10,
                "wrong-integer",
                "metadata is not valid: 'name' metadata property validation failed because 'wrong-integer' is not an integer.",
            ),
            (
                FloatMetadataPropertyFactory,
                13.3,
                "wrong-float",
                "metadata is not valid: 'name' metadata property validation failed because 'wrong-float' is not a float.",
            ),
        ],
    )
    async def test_update_record_with_not_valid_metadata(
        self,
        async_client: "AsyncClient",
        owner_auth_header: dict,
        MetadataPropertyFactoryClass: Type[MetadataPropertyFactory],
        create_value: Any,
        update_value: Any,
        expected_error: str,
    ):
        dataset = await DatasetFactory.create(allow_extra_metadata=False)
        await MetadataPropertyFactoryClass.create(name="name", dataset=dataset)
        record = await RecordFactory.create(dataset=dataset, metadata_={"name": create_value})

        response = await async_client.patch(
            f"/api/v1/records/{record.id}",
            headers=owner_auth_header,
            json={"metadata": {"name": update_value}},
        )

        assert response.status_code == 422
        assert response.json() == {"detail": expected_error}

    async def test_update_record_with_extra_metadata_not_allowed(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create(allow_extra_metadata=False)
        await TermsMetadataPropertyFactory.create(name="terms-metadata-property", dataset=dataset)
        await IntegerMetadataPropertyFactory.create(name="integer-metadata-property", dataset=dataset)
        await FloatMetadataPropertyFactory.create(name="float-metadata-property", dataset=dataset)
        record = await RecordFactory.create(
            dataset=dataset,
            metadata_={"terms-metadata-property": "a", "integer-metadata-property": 1, "float-metadata-property": 1.0},
        )

        response = await async_client.patch(
            f"/api/v1/records/{record.id}",
            headers=owner_auth_header,
            json={
                "metadata": {
                    "terms-metadata-property": "c",
                    "integer-metadata-property": 9999,
                    "float-metadata-property": 9999.0,
                    "extra-metadata": "yes",
                },
            },
        )

        assert response.status_code == 422
        assert response.json() == {
            "detail": "metadata is not valid: 'extra-metadata' metadata property does not exists for dataset "
            f"'{dataset.id}' and extra metadata is not allowed for this dataset"
        }

    async def test_update_record_with_invalid_suggestion(self, async_client: "AsyncClient", owner_auth_header: dict):
        dataset = await DatasetFactory.create()
        question = await LabelSelectionQuestionFactory.create(dataset=dataset)
        record = await RecordFactory.create(dataset=dataset)

        response = await async_client.patch(
            f"/api/v1/records/{record.id}",
            headers=owner_auth_header,
            json={
                "suggestions": [
                    {"question_id": str(question.id), "value": "not a valid value"},
                ]
            },
        )

        assert response.status_code == 422
        assert response.json() == {
            "detail": f"suggestion for question_id={question.id} is not valid: 'not a valid value' is not a valid option.\nValid options are: ['option1', 'option2', 'option3']"
        }

    async def test_update_record_with_invalid_vector(self, async_client: "AsyncClient", owner_auth_header: dict):
        dataset = await DatasetFactory.create()
        vector_settings = await VectorSettingsFactory.create(dataset=dataset, dimensions=5)
        record = await RecordFactory.create(dataset=dataset)

        response = await async_client.patch(
            f"/api/v1/records/{record.id}",
            headers=owner_auth_header,
            json={"vectors": {vector_settings.name: [1, 2, 3, 4, 5, 6]}},
        )

        assert response.status_code == 422
        assert response.json() == {
            "detail": f"vector with name={vector_settings.name} is not valid: vector must have 5 elements, got 6 elements"
        }

    async def test_update_record_with_suggestion_for_nonexistent_question(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()
        record = await RecordFactory.create(dataset=dataset)

        question_id = uuid4()

        response = await async_client.patch(
            f"/api/v1/records/{record.id}",
            headers=owner_auth_header,
            json={
                "suggestions": [
                    {"question_id": str(question_id), "value": "option-1"},
                ]
            },
        )

        assert response.status_code == 422
        assert response.json() == {
            "detail": f"suggestion for question_id={question_id} is not valid: question_id={question_id} does not exist"
        }

    async def test_update_record_with_nonexistent_vector_settings(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()
        record = await RecordFactory.create(dataset=dataset)

        response = await async_client.patch(
            f"/api/v1/records/{record.id}",
            headers=owner_auth_header,
            json={"vectors": {"i-do-not-exist": [1, 2, 3, 4, 5, 6]}},
        )

        assert response.status_code == 422
        assert response.json() == {
            "detail": f"vector with name=i-do-not-exist is not valid: vector with name=i-do-not-exist does not exist for dataset_id={dataset.id}"
        }

    async def test_update_record_with_duplicate_suggestions_question_ids(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()
        question = await TextQuestionFactory.create(dataset=dataset)
        record = await RecordFactory.create(dataset=dataset)

        response = await async_client.patch(
            f"/api/v1/records/{record.id}",
            headers=owner_auth_header,
            json={
                "suggestions": [
                    {"question_id": str(question.id), "value": "a"},
                    {"question_id": str(question.id), "value": "b"},
                ]
            },
        )

        assert response.status_code == 422
        assert response.json() == {"detail": "found duplicate suggestions question IDs"}

    async def test_update_record_as_admin_from_another_workspace(self, async_client: "AsyncClient"):
        record = await RecordFactory.create()
        user = await UserFactory.create(role=UserRole.admin)

        response = await async_client.patch(
            f"/api/v1/records/{record.id}",
            headers={API_KEY_HEADER_NAME: user.api_key},
            json={
                "metadata": {"new": "yes"},
            },
        )

        assert response.status_code == 403

    async def test_update_record_as_annotator(self, async_client: "AsyncClient"):
        record = await RecordFactory.create()
        user = await UserFactory.create(role=UserRole.annotator, workspaces=[record.dataset.workspace])

        response = await async_client.patch(
            f"/api/v1/records/{record.id}",
            headers={API_KEY_HEADER_NAME: user.api_key},
            json={
                "metadata": {"new": "yes"},
            },
        )

        assert response.status_code == 403

    @pytest.mark.parametrize(
        "response_status", [ResponseStatus.submitted, ResponseStatus.discarded, ResponseStatus.draft]
    )
    async def test_create_record_response_with_required_questions(
        self,
        async_client: "AsyncClient",
        db: "AsyncSession",
        mock_search_engine: SearchEngine,
        owner: User,
        owner_auth_header: dict,
        response_status: ResponseStatus,
    ):
        dataset = await DatasetFactory.create()
        await TextQuestionFactory.create(name="corrected-1", dataset=dataset, required=True)
        await TextQuestionFactory.create(name="corrected-2", dataset=dataset)
        record = await RecordFactory.create(dataset=dataset)

        responses = {"values": {"corrected-1": {"value": "Unit Test 1"}}}
        response_json = {**responses, "status": response_status}
        response = await async_client.post(
            f"/api/v1/records/{record.id}/responses", headers=owner_auth_header, json=response_json
        )

        response_body = response.json()
        assert response.status_code == 201
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 1
        assert await db.get(Response, UUID(response_body["id"]))
        assert response_body == {
            "id": str(UUID(response_body["id"])),
            "values": responses["values"],
            "status": response_status,
            "record_id": str(record.id),
            "user_id": str(owner.id),
            "inserted_at": datetime.fromisoformat(response_body["inserted_at"]).isoformat(),
            "updated_at": datetime.fromisoformat(response_body["updated_at"]).isoformat(),
        }

        response = (await db.execute(select(Response).where(Response.record_id == record.id))).scalar_one()
        mock_search_engine.update_record_response.assert_called_once_with(response)

    async def test_create_submitted_record_response_with_missing_required_questions(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()
        await create_text_questions(dataset)

        record = await RecordFactory.create(dataset=dataset)
        response_json = {
            "values": {"output_ok": {"value": "yes"}},
            "status": "submitted",
        }

        response = await async_client.post(
            f"/api/v1/records/{record.id}/responses", headers=owner_auth_header, json=response_json
        )
        assert response.status_code == 422
        assert response.json() == {"detail": "missing question with name=input_ok"}

    @pytest.mark.parametrize("response_status", [ResponseStatus.discarded, ResponseStatus.draft])
    async def test_create_record_response_with_missing_required_questions(
        self,
        async_client: "AsyncClient",
        db: "AsyncSession",
        mock_search_engine: SearchEngine,
        owner: User,
        owner_auth_header: dict,
        response_status: ResponseStatus,
    ):
        dataset = await DatasetFactory.create()
        await TextQuestionFactory.create(name="corrected-1", dataset=dataset, required=True)
        await TextQuestionFactory.create(name="corrected-2", dataset=dataset)
        record = await RecordFactory.create(dataset=dataset)

        responses = {"values": {"corrected-2": {"value": "Unit Test 2"}}}
        response_json = {**responses, "status": response_status}
        response = await async_client.post(
            f"/api/v1/records/{record.id}/responses", headers=owner_auth_header, json=response_json
        )

        response_body = response.json()
        assert response.status_code == 201
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 1
        assert await db.get(Response, UUID(response_body["id"]))
        assert response_body == {
            "id": str(UUID(response_body["id"])),
            "values": responses["values"],
            "status": response_status,
            "record_id": str(record.id),
            "user_id": str(owner.id),
            "inserted_at": datetime.fromisoformat(response_body["inserted_at"]).isoformat(),
            "updated_at": datetime.fromisoformat(response_body["updated_at"]).isoformat(),
        }

        response = (await db.execute(select(Response).where(Response.record_id == record.id))).scalar_one()
        mock_search_engine.update_record_response.assert_called_once_with(response)

    @pytest.mark.parametrize(
        "QuestionFactory, response_value",
        [
            (TextQuestionFactory, "Unit Test!"),
            (RatingQuestionFactory, 3),
            (LabelSelectionQuestionFactory, "option1"),
            (MultiLabelSelectionQuestionFactory, ["option1", "option2"]),
            (
                RankingQuestionFactory,
                [
                    {"value": "completion-a", "rank": 1},
                    {"value": "completion-b", "rank": 2},
                    {"value": "completion-c", "rank": 3},
                ],
            ),
        ],
    )
    async def test_create_record_response_with_submitted_status(
        self,
        async_client: "AsyncClient",
        db: "AsyncSession",
        QuestionFactory: Type[QuestionFactory],
        response_value: Any,
        owner: User,
        owner_auth_header: dict,
    ):
        question = await QuestionFactory.create()
        record = await RecordFactory.create(dataset=question.dataset)

        response_json = {"values": {question.name: {"value": response_value}}, "status": ResponseStatus.submitted}

        response = await async_client.post(
            f"/api/v1/records/{record.id}/responses", headers=owner_auth_header, json=response_json
        )

        response_body = response.json()
        assert response.status_code == 201
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 1
        assert response.json() == {
            "id": str(UUID(response_body["id"])),
            "values": {question.name: {"value": response_value}},
            "status": ResponseStatus.submitted,
            "record_id": str(record.id),
            "user_id": str(owner.id),
            "inserted_at": datetime.fromisoformat(response_body["inserted_at"]).isoformat(),
            "updated_at": datetime.fromisoformat(response_body["updated_at"]).isoformat(),
        }

    @pytest.mark.parametrize("response_status", [ResponseStatus.discarded, ResponseStatus.draft])
    @pytest.mark.parametrize(
        "QuestionFactory, response_value",
        [
            (TextQuestionFactory, "Unit Test!"),
            (RatingQuestionFactory, 3),
            (LabelSelectionQuestionFactory, "option1"),
            (MultiLabelSelectionQuestionFactory, ["option1", "option2"]),
            (
                RankingQuestionFactory,
                [
                    {"value": "completion-a", "rank": 1},
                    {"value": "completion-b", "rank": 2},
                    {"value": "completion-c", "rank": 3},
                ],
            ),
            (
                RankingQuestionFactory,
                [
                    {"value": "completion-a", "rank": 1},
                ],
            ),
        ],
    )
    async def test_create_record_response_with_non_submitted_status(
        self,
        async_client: "AsyncClient",
        db: "AsyncSession",
        response_status: ResponseStatus,
        QuestionFactory: Type[QuestionFactory],
        response_value: Any,
        owner: User,
        owner_auth_header: dict,
    ):
        question = await QuestionFactory.create()
        record = await RecordFactory.create(dataset=question.dataset)

        response_json = {"values": {question.name: {"value": response_value}}, "status": response_status}

        response = await async_client.post(
            f"/api/v1/records/{record.id}/responses", headers=owner_auth_header, json=response_json
        )

        response_body = response.json()
        assert response.status_code == 201
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 1
        assert response.json() == {
            "id": str(UUID(response_body["id"])),
            "values": {question.name: {"value": response_value}},
            "status": response_status.value,
            "record_id": str(record.id),
            "user_id": str(owner.id),
            "inserted_at": datetime.fromisoformat(response_body["inserted_at"]).isoformat(),
            "updated_at": datetime.fromisoformat(response_body["updated_at"]).isoformat(),
        }

    async def test_create_record_response_with_extra_question_responses(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()
        await create_text_questions(dataset)
        record = await RecordFactory.create(dataset=dataset)

        response_json = {
            "values": {
                "input_ok": {"value": "yes"},
                "unknown_question": {"value": "Test"},
            },
            "status": "submitted",
        }
        response = await async_client.post(
            f"/api/v1/records/{record.id}/responses", headers=owner_auth_header, json=response_json
        )

        assert response.status_code == 422
        assert response.json() == {"detail": "found responses for non configured questions: ['unknown_question']"}

    @pytest.mark.parametrize(
        "create_questions_func, responses, expected_error_msg",
        [
            (
                create_text_questions,
                {
                    "values": {
                        "input_ok": {"value": True},
                        "output_ok": {"value": False},
                    },
                },
                "Expected text value, found <class 'bool'>",
            ),
            (
                create_rating_questions,
                {
                    "values": {
                        "rating_question_1": {"value": "wrong-rating-value"},
                    },
                },
                "'wrong-rating-value' is not a valid option.\nValid options are: [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]",
            ),
            (
                create_label_selection_questions,
                {
                    "values": {
                        "label_selection_question_1": {"value": False},
                    },
                },
                "False is not a valid option.\nValid options are: ['option1', 'option2', 'option3']",
            ),
            (
                create_multi_label_selection_questions,
                {
                    "values": {
                        "multi_label_selection_question_1": {"value": "wrong-type"},
                    },
                },
                "This MultiLabelSelection question expects a list of values, found <class 'str'>",
            ),
            (
                create_multi_label_selection_questions,
                {
                    "values": {
                        "multi_label_selection_question_1": {"value": ["option4", "option5"]},
                    },
                },
                "['option4', 'option5'] are not valid options for this MultiLabelSelection question.\nValid options are: ['option1', 'option2', 'option3']",
            ),
            (
                create_multi_label_selection_questions,
                {"values": {"multi_label_selection_question_1": {"value": []}}},
                "This MultiLabelSelection question expects a list of values, found empty list",
            ),
            (
                create_ranking_question,
                {"values": {"ranking_question_1": {"value": "wrong-type"}}},
                "This Ranking question expects a list of values, found <class 'str'>",
            ),
            (
                create_ranking_question,
                {"values": {"ranking_question_1": {"value": []}}},
                "This Ranking question expects a list containing 3 values, found a list of 0 values",
            ),
            (
                create_ranking_question,
                {
                    "values": {
                        "ranking_question_1": {
                            "value": [
                                {"value": "completion-b", "rank": 1},
                            ]
                        }
                    }
                },
                "This Ranking question expects a list containing 3 values, found a list of 1 values",
            ),
            (
                create_ranking_question,
                {
                    "values": {
                        "ranking_question_1": {
                            "value": [
                                {"value": "completion-b", "rank": 1},
                                {"value": "completion-c", "rank": 2},
                                {"value": "completion-a", "rank": 3},
                                {"value": "completion-z", "rank": 4},
                            ],
                        }
                    }
                },
                "This Ranking question expects a list containing 3 values, found a list of 4 values",
            ),
            (
                create_ranking_question,
                {
                    "values": {
                        "ranking_question_1": {
                            "value": [
                                {"value": "completion-b", "rank": 1},
                                {"value": "completion-c", "rank": 2},
                                {"value": "completion-a", "rank": 4},
                            ]
                        }
                    }
                },
                "[4] are not valid ranks for this Ranking question.\nValid ranks are: [1, 2, 3]",
            ),
            (
                create_ranking_question,
                {
                    "values": {
                        "ranking_question_1": {
                            "value": [
                                {"value": "completion-b"},
                                {"value": "completion-c"},
                                {"value": "completion-a"},
                            ]
                        }
                    }
                },
                "[None] are not valid ranks for this Ranking question.\nValid ranks are: [1, 2, 3]",
            ),
            (
                create_ranking_question,
                {
                    "values": {
                        "ranking_question_1": {
                            "value": [
                                {"value": "completion-z", "rank": 1},
                                {"value": "completion-c", "rank": 2},
                                {"value": "completion-a", "rank": 3},
                            ]
                        }
                    }
                },
                "['completion-z'] are not valid options for this Ranking question.\nValid options are: ['completion-a', 'completion-b', 'completion-c']",
            ),
            (
                create_ranking_question,
                {
                    "values": {
                        "ranking_question_1": {
                            "value": [
                                {"value": "completion-a", "rank": 1},
                                {"value": "completion-c", "rank": 2},
                                {"value": "completion-a", "rank": 3},
                            ]
                        }
                    }
                },
                "This Ranking question expects a list of unique values, but duplicates were found",
            ),
        ],
    )
    async def test_create_record_response_with_wrong_response_value(
        self,
        async_client: "AsyncClient",
        owner_auth_header: dict,
        create_questions_func: Callable[["Dataset"], None],
        responses: dict,
        expected_error_msg: str,
    ):
        dataset = await DatasetFactory.create()
        await create_questions_func(dataset)
        record = await RecordFactory.create(dataset=dataset)

        response_json = {**responses, "status": "submitted"}
        response = await async_client.post(
            f"/api/v1/records/{record.id}/responses", headers=owner_auth_header, json=response_json
        )

        assert response.status_code == 422
        assert response.json() == {"detail": expected_error_msg}

    async def test_create_record_response_without_authentication(self, async_client: "AsyncClient", db: "AsyncSession"):
        record = await RecordFactory.create()
        response_json = {
            "values": {
                "input_ok": {"value": "yes"},
                "output_ok": {"value": "yes"},
            },
            "status": "submitted",
        }

        response = await async_client.post(f"/api/v1/records/{record.id}/responses", json=response_json)

        assert response.status_code == 401
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 0

    @pytest.mark.parametrize("status", ["submitted", "discarded", "draft"])
    async def test_create_record_response(
        self, async_client: "AsyncClient", db: "AsyncSession", owner: User, owner_auth_header: dict, status: str
    ):
        dataset = await DatasetFactory.create()
        await TextQuestionFactory.create(name="input_ok", dataset=dataset)
        await TextQuestionFactory.create(name="output_ok", dataset=dataset)

        record = await RecordFactory.create(dataset=dataset)
        response_json = {
            "values": {
                "input_ok": {"value": "yes"},
                "output_ok": {"value": "yes"},
            },
            "status": status,
        }

        dataset_previous_last_activity_at = dataset.last_activity_at
        dataset_previous_updated_at = dataset.updated_at

        response = await async_client.post(
            f"/api/v1/records/{record.id}/responses", headers=owner_auth_header, json=response_json
        )

        assert response.status_code == 201
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 1

        assert dataset.last_activity_at > dataset_previous_last_activity_at
        assert dataset.updated_at == dataset_previous_updated_at

        response_body = response.json()
        assert await db.get(Response, UUID(response_body["id"]))
        assert response_body == {
            "id": str(UUID(response_body["id"])),
            "values": {
                "input_ok": {"value": "yes"},
                "output_ok": {"value": "yes"},
            },
            "status": status,
            "record_id": str(record.id),
            "user_id": str(owner.id),
            "inserted_at": datetime.fromisoformat(response_body["inserted_at"]).isoformat(),
            "updated_at": datetime.fromisoformat(response_body["updated_at"]).isoformat(),
        }

    @pytest.mark.parametrize(
        "status, expected_status_code, expected_response_count",
        [("submitted", 422, 0), ("discarded", 201, 1), ("draft", 201, 1)],
    )
    async def test_create_record_response_without_values(
        self,
        async_client: "AsyncClient",
        db: "AsyncSession",
        owner: User,
        owner_auth_header: dict,
        status: str,
        expected_status_code: int,
        expected_response_count: int,
    ):
        record = await RecordFactory.create()
        response_json = {"status": status}

        response = await async_client.post(
            f"/api/v1/records/{record.id}/responses", headers=owner_auth_header, json=response_json
        )

        assert response.status_code == expected_status_code
        assert (await db.execute(select(func.count(Response.id)))).scalar() == expected_response_count

        if expected_status_code == 201:
            response_body = response.json()
            assert await db.get(Response, UUID(response_body["id"]))
            assert response_body == {
                "id": str(UUID(response_body["id"])),
                "values": None,
                "status": status,
                "record_id": str(record.id),
                "user_id": str(owner.id),
                "inserted_at": datetime.fromisoformat(response_body["inserted_at"]).isoformat(),
                "updated_at": datetime.fromisoformat(response_body["updated_at"]).isoformat(),
            }

    @pytest.mark.parametrize("status", [ResponseStatus.submitted, ResponseStatus.discarded, ResponseStatus.draft])
    async def test_create_record_response_with_wrong_values(
        self, async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict, status: ResponseStatus
    ):
        record = await RecordFactory.create()
        response_json = {"status": status, "values": {"wrong_question": {"value": "wrong value"}}}

        response = await async_client.post(
            f"/api/v1/records/{record.id}/responses", headers=owner_auth_header, json=response_json
        )

        assert response.status_code == 422
        assert response.json() == {"detail": "found responses for non configured questions: ['wrong_question']"}
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 0

    @pytest.mark.parametrize("role", [UserRole.owner, UserRole.admin, UserRole.annotator])
    async def test_create_record_response_for_user_role(self, async_client: "AsyncClient", db: Session, role: UserRole):
        dataset = await DatasetFactory.create()
        await TextQuestionFactory.create(name="input_ok", dataset=dataset)
        await TextQuestionFactory.create(name="output_ok", dataset=dataset)

        record = await RecordFactory.create(dataset=dataset)
        user = await UserFactory.create(workspaces=[record.dataset.workspace], role=role)
        response_json = {
            "values": {
                "input_ok": {"value": "yes"},
                "output_ok": {"value": "yes"},
            },
            "status": "submitted",
        }

        response = await async_client.post(
            f"/api/v1/records/{record.id}/responses", headers={API_KEY_HEADER_NAME: user.api_key}, json=response_json
        )

        assert response.status_code == 201
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 1

        response_body = response.json()
        assert response_body == {
            "id": str(UUID(response_body["id"])),
            "values": {
                "input_ok": {"value": "yes"},
                "output_ok": {"value": "yes"},
            },
            "status": "submitted",
            "record_id": str(record.id),
            "user_id": str(user.id),
            "inserted_at": datetime.fromisoformat(response_body["inserted_at"]).isoformat(),
            "updated_at": datetime.fromisoformat(response_body["updated_at"]).isoformat(),
        }

    @pytest.mark.parametrize("role", [UserRole.admin, UserRole.annotator])
    async def test_create_record_response_as_restricted_user_from_different_workspace(
        self, async_client: "AsyncClient", db: Session, role: UserRole
    ):
        record = await RecordFactory.create()
        workspace = await WorkspaceFactory.create()
        user = await UserFactory.create(workspaces=[workspace], role=role)
        response_json = {
            "values": {
                "input_ok": {"value": "yes"},
                "output_ok": {"value": "yes"},
            },
            "status": "submitted",
        }

        response = await async_client.post(
            f"/api/v1/records/{record.id}/responses", headers={API_KEY_HEADER_NAME: user.api_key}, json=response_json
        )

        assert response.status_code == 403
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 0

    async def test_create_record_response_already_created(
        self, async_client: "AsyncClient", db: "AsyncSession", owner: User, owner_auth_header: dict
    ):
        record = await RecordFactory.create()
        await ResponseFactory.create(record=record, user=owner)
        response_json = {
            "values": {
                "input_ok": {"value": "yes"},
                "output_ok": {"value": "yes"},
            },
            "status": "submitted",
        }

        response = await async_client.post(
            f"/api/v1/records/{record.id}/responses", headers=owner_auth_header, json=response_json
        )

        assert response.status_code == 409
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 1

    async def test_create_record_response_with_invalid_values(
        self, async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
    ):
        record = await RecordFactory.create()
        response_json = {
            "values": "invalid",
            "status": "submitted",
        }

        response = await async_client.post(
            f"/api/v1/records/{record.id}/responses", headers=owner_auth_header, json=response_json
        )

        assert response.status_code == 422
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 0

    async def test_create_record_response_with_invalid_status(
        self, async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
    ):
        record = await RecordFactory.create()
        response_json = {
            "values": {
                "input_ok": {"value": "yes"},
                "output_ok": {"value": "yes"},
            },
            "status": "invalid",
        }

        response = await async_client.post(
            f"/api/v1/records/{record.id}/responses", headers=owner_auth_header, json=response_json
        )

        assert response.status_code == 422
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 0

    async def test_create_record_response_with_nonexistent_record_id(
        self, async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
    ):
        await RecordFactory.create()
        response_json = {
            "values": {
                "input_ok": {"value": "yes"},
                "output_ok": {"value": "yes"},
            },
            "status": "submitted",
        }

        response = await async_client.post(
            f"/api/v1/records/{uuid4()}/responses", headers=owner_auth_header, json=response_json
        )

        assert response.status_code == 404
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 0

    @pytest.mark.parametrize("role", [UserRole.annotator, UserRole.admin, UserRole.owner])
    async def test_get_record_suggestions(self, async_client: "AsyncClient", role: UserRole):
        dataset = await DatasetFactory.create()
        user = await UserFactory.create(role=role, workspaces=[dataset.workspace])
        record = await RecordFactory.create(dataset=dataset)
        question_a = await TextQuestionFactory.create(dataset=dataset)
        question_b = await TextQuestionFactory.create(dataset=dataset)
        suggestion_a = await SuggestionFactory.create(
            question=question_a, record=record, value="This is a unit test suggestion"
        )
        suggestion_b = await SuggestionFactory.create(
            question=question_b, record=record, value="This is a another unit test suggestion"
        )

        response = await async_client.get(
            f"/api/v1/records/{record.id}/suggestions", headers={API_KEY_HEADER_NAME: user.api_key}
        )

        assert response.status_code == 200
        assert response.json() == {
            "items": [
                {
                    "id": str(suggestion_a.id),
                    "question_id": str(question_a.id),
                    "type": None,
                    "score": None,
                    "value": "This is a unit test suggestion",
                    "agent": None,
                },
                {
                    "id": str(suggestion_b.id),
                    "question_id": str(question_b.id),
                    "type": None,
                    "score": None,
                    "value": "This is a another unit test suggestion",
                    "agent": None,
                },
            ]
        }

    @pytest.mark.parametrize(
        "payload",
        [
            {
                "type": "model",
                "score": 1,
                "value": "This is a unit test suggestion",
                "agent": "unit-test-agent",
            },
            {
                "type": None,
                "score": None,
                "value": "This is a unit test suggestion",
                "agent": None,
            },
        ],
    )
    @pytest.mark.parametrize("role", [UserRole.admin, UserRole.owner])
    async def test_create_record_suggestion(
        self, async_client: "AsyncClient", db: "AsyncSession", role: UserRole, payload: dict
    ):
        dataset = await DatasetFactory.create()
        question = await TextQuestionFactory.create(dataset=dataset)
        user = await UserFactory.create(role=role, workspaces=[dataset.workspace])
        record = await RecordFactory.create(dataset=dataset)

        response = await async_client.put(
            f"/api/v1/records/{record.id}/suggestions",
            headers={API_KEY_HEADER_NAME: user.api_key},
            json={"question_id": str(question.id), **payload},
        )

        response_body = response.json()
        assert response.status_code == 201
        assert response_body == {"id": response_body["id"], "question_id": str(question.id), **payload}
        assert (await db.execute(select(func.count(Suggestion.id)))).scalar() == 1

    async def test_create_record_suggestion_update(
        self, async_client: "AsyncClient", db: "AsyncSession", mock_search_engine: SearchEngine, owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()
        question = await TextQuestionFactory.create(dataset=dataset)
        record = await RecordFactory.create(dataset=dataset)
        suggestion = await SuggestionFactory.create(question=question, record=record)

        response = await async_client.put(
            f"/api/v1/records/{record.id}/suggestions",
            headers=owner_auth_header,
            json={"question_id": str(question.id), "value": "Testing updating a suggestion"},
        )

        response_body = response.json()
        assert response.status_code == 200
        assert response_body == {
            "id": str(suggestion.id),
            "question_id": str(question.id),
            "type": None,
            "score": None,
            "value": "Testing updating a suggestion",
            "agent": None,
        }
        assert (await db.execute(select(func.count(Suggestion.id)))).scalar() == 1

        mock_search_engine.update_record_suggestion.assert_called_once_with(suggestion)

    @pytest.mark.parametrize(
        "payload",
        [
            {},  # missing value
            {
                "value": {"this": "is not valid response for a TextQuestion"},
            },
        ],
    )
    async def test_create_record_suggestion_not_valid(
        self, async_client: "AsyncClient", owner_auth_header: dict, payload: dict
    ):
        dataset = await DatasetFactory.create()
        question = await TextQuestionFactory.create(dataset=dataset)
        record = await RecordFactory.create(dataset=dataset)

        response = await async_client.put(
            f"/api/v1/records/{record.id}/suggestions",
            headers=owner_auth_header,
            json={"question_id": str(question.id), **payload},
        )

        assert response.status_code == 422

    async def test_create_record_suggestion_for_non_existent_question(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        record = await RecordFactory.create()

        response = await async_client.put(
            f"/api/v1/records/{record.id}/suggestions",
            headers=owner_auth_header,
            json={"question_id": str(uuid4()), "value": "This is a unit test suggestion"},
        )

        assert response.status_code == 422

    async def test_create_record_suggestion_as_annotator(self, async_client: "AsyncClient"):
        annotator = await UserFactory.create(role=UserRole.annotator)
        record = await RecordFactory.create()

        response = await async_client.put(
            f"/api/v1/records/{record.id}/suggestions",
            headers={API_KEY_HEADER_NAME: annotator.api_key},
            json={"question_id": str(uuid4()), "value": "This is a unit test suggestion"},
        )

        assert response.status_code == 403

    @pytest.mark.parametrize("role", [UserRole.owner, UserRole.admin])
    async def test_delete_record(
        self, async_client: "AsyncClient", db: "AsyncSession", mock_search_engine: "SearchEngine", role: UserRole
    ):
        dataset = await DatasetFactory.create()
        record = await RecordFactory.create(dataset=dataset)
        user = await UserFactory.create(role=role, workspaces=[dataset.workspace])

        response = await async_client.delete(
            f"/api/v1/records/{record.id}", headers={API_KEY_HEADER_NAME: user.api_key}
        )

        assert response.status_code == 200
        assert response.json() == {
            "id": str(record.id),
            "fields": record.fields,
            "metadata": None,
            "external_id": record.external_id,
            "dataset_id": str(record.dataset_id),
            "inserted_at": record.inserted_at.isoformat(),
            "updated_at": record.updated_at.isoformat(),
        }
        assert (await db.execute(select(func.count(Record.id)))).scalar() == 0
        mock_search_engine.delete_records.assert_called_once_with(dataset=dataset, records=[record])

    async def test_delete_record_as_admin_from_another_workspace(self, async_client: "AsyncClient", db: "AsyncSession"):
        dataset = await DatasetFactory.create()
        record = await RecordFactory.create(dataset=dataset)
        user = await UserFactory.create(role=UserRole.admin)

        response = await async_client.delete(
            f"/api/v1/records/{record.id}", headers={API_KEY_HEADER_NAME: user.api_key}
        )

        assert response.status_code == 403
        assert (await db.execute(select(func.count(Record.id)))).scalar() == 1

    async def test_delete_record_as_annotator(self, async_client: "AsyncClient"):
        annotator = await UserFactory.create(role=UserRole.annotator)
        record = await RecordFactory.create()

        response = await async_client.delete(
            f"/api/v1/records/{record.id}", headers={API_KEY_HEADER_NAME: annotator.api_key}
        )

        assert response.status_code == 403

    async def test_delete_record_non_existent(self, async_client: "AsyncClient", owner_auth_header: dict):
        response = await async_client.delete(f"/api/v1/records/{uuid4()}", headers=owner_auth_header)
        assert response.status_code == 404

    @pytest.mark.parametrize("role", [UserRole.admin, UserRole.owner])
    async def test_delete_record_suggestions(
        self, async_client: "AsyncClient", db: "AsyncSession", mock_search_engine: SearchEngine, role: UserRole
    ) -> None:
        dataset = await DatasetFactory.create()
        user = await UserFactory.create(workspaces=[dataset.workspace], role=role)
        record = await RecordFactory.create(dataset=dataset)
        suggestions = await SuggestionFactory.create_batch(10, record=record)
        random_uuids = [str(uuid4()) for _ in range(0, 5)]

        suggestions_ids = [str(suggestion.id) for suggestion in suggestions]

        uuids_str = ",".join(suggestions_ids + random_uuids)

        response = await async_client.delete(
            f"/api/v1/records/{record.id}/suggestions",
            headers={API_KEY_HEADER_NAME: user.api_key},
            params={"ids": uuids_str},
        )

        assert response.status_code == 204
        assert (await db.execute(select(func.count(Suggestion.id)))).scalar() == 0

        expected_calls = [call(suggestion) for suggestion in suggestions]
        mock_search_engine.delete_record_suggestion.assert_has_calls(expected_calls)

    async def test_delete_record_suggestions_with_no_ids(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ) -> None:
        record = await RecordFactory.create()

        response = await async_client.delete(
            f"/api/v1/records/{record.id}/suggestions",
            headers=owner_auth_header,
            params={"ids": ""},
        )

        assert response.status_code == 422

    async def test_delete_record_suggestions_exceeding_limit(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ) -> None:
        record = await RecordFactory.create()
        suggestions = await SuggestionFactory.create_batch(200, record=record)

        suggestions_ids = [str(suggestion.id) for suggestion in suggestions]

        response = await async_client.delete(
            f"/api/v1/records/{record.id}/suggestions",
            headers=owner_auth_header,
            params={"ids": ",".join(suggestions_ids)},
        )

        assert response.status_code == 422

    async def test_delete_record_suggestions_from_another_record(
        self, async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
    ) -> None:
        record_a = await RecordFactory.create()
        record_b = await RecordFactory.create()
        suggestions_a = await SuggestionFactory.create_batch(10, record=record_a)
        suggestions_b = await SuggestionFactory.create_batch(10, record=record_b)

        suggestions_a_ids = [str(suggestion.id) for suggestion in suggestions_a]
        suggestions_b_ids = [str(suggestion.id) for suggestion in suggestions_b]

        uuids_str = ",".join(suggestions_a_ids + suggestions_b_ids)

        response = await async_client.delete(
            f"/api/v1/records/{record_a.id}/suggestions",
            headers=owner_auth_header,
            params={"ids": uuids_str},
        )

        assert response.status_code == 204
        assert (await db.execute(select(func.count(Suggestion.id)))).scalar() == 10

    async def test_delete_record_suggestions_as_admin_from_another_workspace(self, async_client: "AsyncClient") -> None:
        record = await RecordFactory.create()
        suggestions = await SuggestionFactory.create_batch(10, record=record)
        user = await UserFactory.create(role=UserRole.admin)

        response = await async_client.delete(
            f"/api/v1/records/{record.id}/suggestions",
            headers={API_KEY_HEADER_NAME: user.api_key},
            params={"ids": ",".join([str(suggestion.id) for suggestion in suggestions])},
        )

        assert response.status_code == 403

    async def test_delete_record_suggestions_as_annotator(
        self, async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
    ) -> None:
        record = await RecordFactory.create()
        user = await UserFactory.create(role=UserRole.annotator, workspaces=[record.dataset.workspace])

        response = await async_client.delete(
            f"/api/v1/records/{record.id}/suggestions",
            headers={API_KEY_HEADER_NAME: user.api_key},
            params={"ids": ""},
        )

        assert response.status_code == 403



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os
from typing import TYPE_CHECKING, Optional
from uuid import uuid4

from argilla_server.pydantic_v1 import BaseSettings, PrivateAttr, validator

if TYPE_CHECKING:
    from argilla_server.security.authentication.oauth2 import OAuth2Settings


class Settings(BaseSettings):
    """
    Attributes
    ----------

    secret_key:
        The secret key used for signed the token data

    algorithm:
        Encryption algorithm for token data

    token_expiration:
        The session token expiration . Default=86400 (1 day)

    oauth_cfg:
        The path to the oauth yaml configuration file. Default=.oauth.yaml

    """

    secret_key: str = uuid4().hex
    algorithm: str = "HS256"
    token_expiration: int = 24 * 60 * 60  # 1 day
    oauth_cfg: str = ".oauth.yaml"

    _oauth_settings: Optional["OAuth2Settings"] = PrivateAttr(None)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._oauth_settings = None

    @validator("token_expiration", always=True)
    def default_token_expiration(cls, v, values, **kwargs) -> int:
        if "token_expiration" in values:
            return v

        # This is a backwards compatibility hack to support the old env variable and
        # it will be removed in version 1.25.0. See https://github.com/argilla-io/argilla/issues/4542
        expiration_in_minutes = os.getenv("ARGILLA_LOCAL_AUTH_TOKEN_EXPIRATION_IN_MINUTES")
        if expiration_in_minutes is not None:
            return int(expiration_in_minutes) * 60

        return v

    @property
    def oauth(self) -> "OAuth2Settings":
        """Return the oauth settings"""
        from argilla_server.security.authentication.oauth2.settings import OAuth2Settings

        if self._oauth_settings:
            return self._oauth_settings

        if not self._oauth_settings and os.path.exists(self.oauth_cfg):
            self._oauth_settings = OAuth2Settings.from_yaml(self.oauth_cfg)
        else:
            self._oauth_settings = OAuth2Settings(enabled=False)

        return self._oauth_settings

    class Config:
        env_prefix = "ARGILLA_AUTH_"

        fields = {
            # Support for old local auth env variables.
            # It will be removed in version 1.25.0 (See https://github.com/argilla-io/argilla/issues/4542)
            "algorithm": {"env": ["ARGILLA_LOCAL_AUTH_ALGORITHM", f"{env_prefix}ALGORITHM"]},
            "secret_key": {"env": ["ARGILLA_LOCAL_AUTH_SECRET_KEY", f"{env_prefix}SECRET_KEY"]},
            "token_expiration_in_minutes": {
                "env": [
                    "ARGILLA_LOCAL_AUTH_TOKEN_EXPIRATION_IN_MINUTES",
                    f"{env_prefix}TOKEN_EXPIRATION_IN_MINUTES",
                ]
            },
        }


settings = Settings()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import dataclasses
from typing import Any, Dict

from argilla_server.daos.backend.metrics.base import (
    BidimensionalTermsAggregation,
    HistogramAggregation,
    NestedBidimensionalTermsAggregation,
    NestedHistogramAggregation,
    NestedPathElasticsearchMetric,
    NestedTermsAggregation,
    TermsAggregation,
)
from argilla_server.daos.backend.query_helpers import aggregations

_DEFAULT_MAX_ENTITY_BUCKET = 1000


@dataclasses.dataclass
class TopKMentionsConsistency(NestedPathElasticsearchMetric):
    """Computes the entity consistency distribution"""

    mention_field: str
    labels_field: str

    def _inner_aggregation(
        self,
        size: int,
        interval: int = 1,
        entity_size: int = _DEFAULT_MAX_ENTITY_BUCKET,
    ) -> Dict[str, Any]:
        size = size or 50
        interval = interval or 1
        return {
            "consistency": {
                **aggregations.terms_aggregation(
                    self.compound_nested_field(self.mention_field),
                    size=size,
                ),
                "aggs": {
                    "entities": aggregations.terms_aggregation(
                        self.compound_nested_field(self.labels_field),
                        size=entity_size,
                    ),
                    "count": {"cardinality": {"field": self.compound_nested_field(self.labels_field)}},
                    "entities_variability_filter": {
                        "bucket_selector": {
                            "buckets_path": {"numLabels": "count"},
                            "script": f"params.numLabels >= {interval}",
                        }
                    },
                    "sortby_entities_count": {
                        "bucket_sort": {
                            "sort": [{"count": {"order": "desc"}}],
                            "size": size,
                        }
                    },
                },
            }
        }

    def aggregation_result(self, aggregation_result: Dict[str, Any]) -> Dict[str, Any]:
        """Simplifies the aggregation result sorting by worst mention consistency"""
        result = [
            {
                "mention": mention,
                "entities": [{"label": entity, "count": count} for entity, count in mention_aggs["entities"].items()],
            }
            for mention, mention_aggs in aggregation_result.items()
        ]
        # TODO: filter by entities threshold
        result.sort(key=lambda m: len(m["entities"]), reverse=True)
        return {"mentions": result}


METRICS = {
    "token_frequency": NestedTermsAggregation(
        id="token_frequency",
        nested_path="metrics.tokens",
        terms=TermsAggregation(
            id="frequency",
            field="value",
        ),
    ),
    "token_capitalness": NestedTermsAggregation(
        id="token_capitalness",
        nested_path="metrics.tokens",
        terms=TermsAggregation(
            id="capitalness",
            field="capitalness",
        ),
    ),
    "predicted_entity_labels": NestedTermsAggregation(
        id="predicted_entity_labels",
        nested_path="metrics.predicted.mentions",
        terms=TermsAggregation(
            id="terms",
            field="label",
            default_size=_DEFAULT_MAX_ENTITY_BUCKET,
        ),
    ),
    "annotated_entity_labels": NestedTermsAggregation(
        id="annotated_entity_labels",
        nested_path="metrics.annotated.mentions",
        terms=TermsAggregation(
            id="terms",
            field="label",
            default_size=_DEFAULT_MAX_ENTITY_BUCKET,
        ),
    ),
    "predicted_entity_capitalness": NestedTermsAggregation(
        id="predicted_entity_capitalness",
        nested_path="metrics.predicted.mentions",
        terms=TermsAggregation(id="terms", field="capitalness"),
    ),
    "annotated_entity_capitalness": NestedTermsAggregation(
        id="annotated_entity_capitalness",
        nested_path="metrics.annotated.mentions",
        terms=TermsAggregation(id="terms", field="capitalness"),
    ),
    "predicted_mentions_distribution": NestedBidimensionalTermsAggregation(
        id="predicted_mentions_distribution",
        nested_path="metrics.predicted.mentions",
        biterms=BidimensionalTermsAggregation(id="bi-dimensional", field_x="label", field_y="value"),
    ),
    "annotated_mentions_distribution": NestedBidimensionalTermsAggregation(
        id="predicted_mentions_distribution",
        nested_path="metrics.annotated.mentions",
        biterms=BidimensionalTermsAggregation(id="bi-dimensional", field_x="label", field_y="value"),
    ),
    "predicted_top_k_mentions_consistency": TopKMentionsConsistency(
        id="predicted_top_k_mentions_consistency",
        nested_path="metrics.predicted.mentions",
        mention_field="value",
        labels_field="label",
    ),
    "annotated_top_k_mentions_consistency": TopKMentionsConsistency(
        id="annotated_top_k_mentions_consistency",
        nested_path="metrics.annotated.mentions",
        mention_field="value",
        labels_field="label",
    ),
    # TODO: Remove in v1.10.0
    "tokens_length": HistogramAggregation(
        "tokens_length",
        field="metrics.tokens_length",
    ),
    # TODO: Remove in v1.10.0
    "token_length": NestedHistogramAggregation(
        id="token_length",
        nested_path="metrics.tokens",
        histogram=HistogramAggregation(
            id="length",
            field="length",
            fixed_interval=1,
        ),
    ),
    # TODO: Remove in v1.10.0
    "predicted_mention_char_length": NestedHistogramAggregation(
        id="predicted_mention_char_length",
        nested_path="metrics.predicted.mentions",
        histogram=HistogramAggregation(
            id="length",
            field="chars_length",
            fixed_interval=1,
        ),
    ),
    # TODO: Remove in v1.10.0
    "annotated_mention_char_length": NestedHistogramAggregation(
        id="annotated_mention_char_length",
        nested_path="metrics.annotated.mentions",
        histogram=HistogramAggregation(
            id="length",
            field="chars_length",
            fixed_interval=1,
        ),
    ),
    # TODO: Remove in v1.10.0
    "predicted_entity_density": NestedHistogramAggregation(
        id="predicted_entity_density",
        nested_path="metrics.predicted.mentions",
        histogram=HistogramAggregation(id="histogram", field="density"),
    ),
    # TODO: Remove in v1.10.0
    "annotated_entity_density": NestedHistogramAggregation(
        id="annotated_entity_density",
        nested_path="metrics.annotated.mentions",
        histogram=HistogramAggregation(id="histogram", field="density"),
    ),
    # TODO: Remove in v1.10.0
    "predicted_mention_token_length": NestedHistogramAggregation(
        id="predicted_mention_token_length",
        nested_path="metrics.predicted.mentions",
        histogram=HistogramAggregation(id="histogram", field="tokens_length"),
    ),
    # TODO: Remove in v1.10.0
    "annotated_mention_token_length": NestedHistogramAggregation(
        id="annotated_mention_token_length",
        nested_path="metrics.annotated.mentions",
        histogram=HistogramAggregation(id="histogram", field="tokens_length"),
    ),
    # TODO: Remove in v1.10.0
    "predicted_tag_consistency": TopKMentionsConsistency(
        id="predicted_tag_consistency",
        nested_path="metrics.predicted.tags",
        mention_field="value",
        labels_field="tag",
    ),
    # TODO: Remove in v1.10.0
    "annotated_tag_consistency": TopKMentionsConsistency(
        id="annotated_tag_consistency",
        nested_path="metrics.annotated.tags",
        mention_field="value",
        labels_field="tag",
    ),
}



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import os
import random
import re
import string
from typing import Any, ClassVar, Dict, List, Optional, Tuple, Type, Union
from urllib.parse import urljoin

import httpx
from fastapi import Request
from fastapi.responses import RedirectResponse
from oauthlib.oauth2 import WebApplicationClient
from social_core.backends.oauth import BaseOAuth2
from social_core.exceptions import AuthException
from social_core.strategy import BaseStrategy

from argilla_server.errors import future
from argilla_server.security.authentication.claims import Claims
from argilla_server.security.settings import settings


class Strategy(BaseStrategy):
    def request_data(self, merge=True) -> Dict[str, Any]:
        return {}

    def absolute_uri(self, path=None) -> str:
        return path

    def get_setting(self, name):
        return None


class OAuth2ClientProvider:
    """OAuth2 flow handler  of a certain provider."""

    OAUTH_STATE_COOKIE_NAME = "oauth2_state"
    OAUTH_STATE_COOKIE_MAX_AGE = 10

    name: ClassVar[str]
    backend_class: ClassVar[Type[BaseOAuth2]]
    claims: ClassVar[Optional[Union[Claims, dict]]] = None
    backend_strategy: ClassVar[BaseStrategy] = Strategy()

    def __init__(
        self,
        client_id: str = None,
        client_secret: str = None,
        scope: Optional[List[str]] = None,
        redirect_uri: str = None,
    ) -> None:
        self.client_id = client_id or self._environment_variable_for_property("client_id")
        self.client_secret = client_secret or self._environment_variable_for_property("client_secret")
        self.scope = scope or self._environment_variable_for_property("scope", "")
        self.scope = self.scope.split(" ") if self.scope else []
        self.redirect_uri = redirect_uri or self._environment_variable_for_property("redirect_uri")
        self.redirect_uri = self.redirect_uri or f"/oauth/{self.name}/callback"

        self._backend = self.backend_class(strategy=self.backend_strategy)
        self._authorization_endpoint = self._backend.authorization_url()
        self._token_endpoint = self._backend.access_token_url()

    @classmethod
    def from_dict(cls, provider: dict) -> "OAuth2ClientProvider":
        return cls(**provider)

    def new_oauth_client(self) -> WebApplicationClient:
        return WebApplicationClient(self.client_id)

    def get_redirect_uri(self, request: Request) -> str:
        url = urljoin(str(request.base_url), self.redirect_uri)
        return self._align_url_to_allow_http_redirect(url)

    def authorization_url(self, request: Request) -> Tuple[str, Optional[str]]:
        redirect_uri = self.get_redirect_uri(request)
        state = "".join([random.choice(string.ascii_letters) for _ in range(32)])

        oauth2_query_params = dict(state=state, scope=self.scope, redirect_uri=redirect_uri)
        oauth2_query_params.update(request.query_params)

        authorization_url = str(
            self.new_oauth_client().prepare_request_uri(self._authorization_endpoint, **oauth2_query_params)
        )

        return authorization_url, state

    def authorization_redirect(self, request: Request) -> RedirectResponse:
        url, state = self.authorization_url(request)
        response = RedirectResponse(url, 303)

        response.set_cookie(
            self.OAUTH_STATE_COOKIE_NAME,
            value=state,
            secure=True,
            httponly=True,
            max_age=self.OAUTH_STATE_COOKIE_MAX_AGE,
            samesite="none",
        )

        return response

    async def get_user_data(self, request: Request) -> dict:
        self._check_request_params(request)

        redirect_uri = self.get_redirect_uri(request)
        authorization_response = self._align_url_to_allow_http_redirect(str(request.url))

        oauth2_query_params = dict(redirect_url=redirect_uri)
        oauth2_query_params.update(request.query_params)

        return await self._fetch_user_data(authorization_response=authorization_response, **oauth2_query_params)

    def _check_request_params(self, request) -> None:
        if "code" not in request.query_params:
            raise ValueError("'code' parameter was not found in callback request")

        if "state" not in request.query_params:
            raise ValueError("'state' parameter was not found in callback request")

        state = request.cookies.get(self.OAUTH_STATE_COOKIE_NAME)
        if request.query_params.get("state") != state:
            raise ValueError("'state' parameter does not match")

    @staticmethod
    def _align_url_to_allow_http_redirect(url: str) -> str:
        """This method is used to align the URL to the HTTP/HTTPS scheme"""

        scheme = "http" if settings.oauth.allow_http_redirect else "https"
        return re.sub(r"^https?", scheme, url)

    def standardize(self, data: Dict[str, Any]) -> Dict[str, Any]:
        data["provider"] = self.name
        data["scope"] = self.scope

        return data

    async def _fetch_user_data(self, authorization_response: str, **oauth_query_params) -> dict:
        oauth_client = self.new_oauth_client()
        token_url, headers, content = oauth_client.prepare_token_request(
            self._token_endpoint,
            authorization_response=authorization_response,
            **oauth_query_params,
        )

        headers.update({"Accept": "application/json"})
        auth = httpx.BasicAuth(self.client_id, self.client_secret)
        async with httpx.AsyncClient(auth=auth) as session:
            try:
                response = await session.post(token_url, headers=headers, content=content)
                oauth_client.parse_request_body_response(json.dumps(response.json()))

                return self.standardize(self._backend.user_data(oauth_client.access_token))
            except httpx.HTTPError as e:
                raise ValueError(str(e))
            except AuthException as e:
                raise future.AuthenticationError(str(e))

    def _environment_variable_for_property(self, property_name: str, default: str = None) -> str:
        env_var_name = f"OAUTH2_{self.name.upper()}_{property_name.upper()}"

        return os.getenv(env_var_name, default)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from social_core.backends.github import GithubOAuth2
from social_core.backends.open_id_connect import OpenIdConnectAuth

from argilla_server.security.authentication.claims import Claims
from argilla_server.security.authentication.oauth2.client_provider import OAuth2ClientProvider


class HuggingfaceOpenId(OpenIdConnectAuth):
    """Huggingface OpenID Connect authentication backend."""

    name = "huggingface"

    OIDC_ENDPOINT = "https://huggingface.co"
    AUTHORIZATION_URL = "https://huggingface.co/oauth/authorize"
    ACCESS_TOKEN_URL = "https://huggingface.co/oauth/token"

    def oidc_endpoint(self) -> str:
        return self.OIDC_ENDPOINT


class GitHubClientProvider(OAuth2ClientProvider):
    claims = Claims(picture="avatar_url", identity=lambda user: f"{user.provider}:{user.id}", username="login")
    backend_class = GithubOAuth2
    name = "github"


class HuggingfaceClientProvider(OAuth2ClientProvider):
    """Specialized HuggingFace OAuth2 provider."""

    claims = Claims(username="preferred_username")
    backend_class = HuggingfaceOpenId
    name = "huggingface"


_providers = [GitHubClientProvider, HuggingfaceClientProvider]

ALL_SUPPORTED_OAUTH2_PROVIDERS = {provider_class.name: provider_class for provider_class in _providers}



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import Iterable, List, Optional, Set, Type, Union

from fastapi import Depends

from argilla_server.daos.models.records import DaoRecordsSearch
from argilla_server.daos.records import DatasetRecordsDAO
from argilla_server.errors import RecordNotFound
from argilla_server.services.datasets import ServiceDataset
from argilla_server.services.metrics import MetricsService
from argilla_server.services.metrics.models import ServiceMetric
from argilla_server.services.search.model import (
    ServiceRecord,
    ServiceRecordsQuery,
    ServiceSearchResults,
    ServiceSortConfig,
)


class SearchRecordsService:
    """Generic service for search records operations"""

    _INSTANCE: "SearchRecordsService" = None

    __LOGGER__ = logging.getLogger(__name__)

    @classmethod
    def get_instance(
        cls,
        dao: DatasetRecordsDAO = Depends(DatasetRecordsDAO.get_instance),
        metrics: MetricsService = Depends(MetricsService.get_instance),
    ):
        if not cls._INSTANCE:
            cls._INSTANCE = cls(dao=dao, metrics=metrics)
        return cls._INSTANCE

    def __init__(
        self,
        dao: DatasetRecordsDAO,
        metrics: MetricsService,
    ):
        self.__dao__ = dao
        self.__metrics__ = metrics

    def search(
        self,
        dataset: ServiceDataset,
        record_type: Type[ServiceRecord],
        query: Optional[ServiceRecordsQuery] = None,
        sort_config: Optional[ServiceSortConfig] = None,
        record_from: int = 0,
        size: int = 100,
        exclude_metrics: bool = True,
        metrics: Optional[List[ServiceMetric]] = None,
    ) -> ServiceSearchResults:
        if record_from > 0:
            metrics = None

        sort_config = sort_config or ServiceSortConfig()
        exclude_fields = ["metrics.*"] if exclude_metrics else None
        if query and query.vector and not query.vector.k:
            query.vector.k = size
        results = self.__dao__.search_records(
            dataset,
            search=DaoRecordsSearch(
                query=query,
                sort=sort_config,
            ),
            size=size,
            record_from=record_from,
            exclude_fields=exclude_fields,
            highligth_results=query is not None and query.query_text is not None and len(query.query_text) > 0,
        )
        metrics_results = {}
        for metric in metrics or []:
            try:
                metrics_ = self.__metrics__.summarize_metric(
                    dataset=dataset,
                    metric=metric,
                    record_class=record_type,
                    query=query,
                )
                metrics_results[metric.id] = metrics_
            except Exception as ex:
                self.__LOGGER__.warning("Cannot compute metric [%s]. Error: %s", metric.id, ex)
                metrics_results[metric.id] = {}

        return ServiceSearchResults(
            total=results.total,
            records=[record_type.parse_obj(r) for r in results.records],
            metrics=metrics_results if metrics_results else {},
        )

    async def find_record_by_id(
        self,
        dataset: ServiceDataset,
        id: Union[str, int],
        record_type: Type[ServiceRecord],
    ) -> ServiceRecord:
        found = await self.__dao__.get_record_by_id(dataset, id)
        if not found:
            raise RecordNotFound(
                dataset=dataset.id,
                id=id,
                type="Record",
            )

        return record_type.parse_obj(found)

    def scan_records(
        self,
        dataset: ServiceDataset,
        record_type: Optional[Type[ServiceRecord]] = None,
        query: Optional[ServiceRecordsQuery] = None,
        projection: Set[str] = None,
        id_from: Optional[str] = None,
        limit: int = 1000,
    ) -> Iterable[ServiceRecord]:
        """Scan records for a queried"""
        search = DaoRecordsSearch(query=query)

        transform_doc = lambda doc: doc
        if record_type:
            transform_doc = record_type.parse_obj

        for doc in self.__dao__.scan_dataset(
            dataset,
            id_from=id_from,
            limit=limit,
            search=search,
            include_fields=projection,
        ):
            yield transform_doc(doc)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import dataclasses
from typing import Any, Dict, List, Optional
from uuid import UUID

from elasticsearch8 import AsyncElasticsearch, helpers

from argilla_server.models import VectorSettings
from argilla_server.search_engine import SearchEngine
from argilla_server.search_engine.commons import (
    BaseElasticAndOpenSearchEngine,
    es_bool_query,
    es_field_for_vector_settings,
    es_ids_query,
)
from argilla_server.settings import settings


def _compute_num_candidates_from_k(k: int) -> int:
    if k < 50:
        return 500
    elif 50 <= k < 200:
        return 100
    return 2000


@SearchEngine.register(engine_name="elasticsearch")
@dataclasses.dataclass
class ElasticSearchEngine(BaseElasticAndOpenSearchEngine):
    config: Dict[str, Any] = dataclasses.field(default_factory=dict)

    def __post_init__(self):
        self.client = AsyncElasticsearch(**self.config)

    @classmethod
    async def new_instance(cls) -> "ElasticSearchEngine":
        config = dict(
            hosts=settings.elasticsearch,
            verify_certs=settings.elasticsearch_ssl_verify,
            ca_certs=settings.elasticsearch_ca_path,
            retry_on_timeout=True,
            max_retries=5,
        )
        return cls(
            config=config,
            number_of_shards=settings.es_records_index_shards,
            number_of_replicas=settings.es_records_index_replicas,
        )

    async def close(self):
        await self.client.close()

    def _configure_index_settings(self) -> Dict[str, Any]:
        return {
            "max_result_window": self.max_result_window,
            "number_of_shards": self.number_of_shards,
            "number_of_replicas": self.number_of_replicas,
        }

    def _mapping_for_vector_settings(self, vector_settings: VectorSettings) -> dict:
        return {
            es_field_for_vector_settings(vector_settings): {
                "type": "dense_vector",
                "dims": vector_settings.dimensions,
                "index": True,
                # can similarity property also be part of config @frascuchon ?
                # relates vector search similarity metric
                # "similarity": "l2_norm",  ## default value regarding the knn best practices es documentation
                "similarity": "cosine",
            }
        }

    async def _request_similarity_search(
        self,
        index: str,
        vector_settings: VectorSettings,
        value: List[float],
        k: int,
        excluded_id: Optional[UUID] = None,
        query_filters: Optional[List[dict]] = None,
    ) -> dict:
        knn_query = {
            "field": es_field_for_vector_settings(vector_settings),
            "query_vector": value,
            "k": k,
            "num_candidates": _compute_num_candidates_from_k(k=k),
        }

        if bool(excluded_id) or bool(query_filters):
            minimum_should_match = None
            must_not_filters = None

            if excluded_id:
                must_not_filters = [es_ids_query([str(excluded_id)])]
            if query_filters:
                minimum_should_match = len(query_filters)

            bool_filter_query = es_bool_query(
                should=query_filters, minimum_should_match=minimum_should_match, must_not=must_not_filters
            )

            knn_query["filter"] = bool_filter_query
        return await self.client.search(index=index, knn=knn_query, _source=False, track_total_hits=True, size=k)

    async def _create_index_request(self, index_name: str, mappings: dict, settings: dict) -> None:
        await self.client.indices.create(index=index_name, settings=settings, mappings=mappings)

    async def _delete_index_request(self, index_name: str):
        await self.client.indices.delete(index=index_name, ignore=[404], ignore_unavailable=True)

    async def _update_document_request(self, index_name: str, id: str, body: dict):
        # https://www.elastic.co/guide/en/elasticsearch/reference/current/indices-refresh.html#refresh-api-desc
        await self.client.update(index=index_name, id=id, **body, refresh=True)

    async def put_index_mapping_request(self, index: str, mappings: dict):
        await self.client.indices.put_mapping(index=index, properties=mappings)

    async def _index_search_request(
        self,
        index: str,
        query: dict,
        size: Optional[int] = None,
        from_: Optional[int] = None,
        sort: str = None,
        aggregations: Optional[dict] = None,
    ) -> dict:
        return await self.client.search(
            index=index,
            query=query,
            from_=from_,
            size=size,
            source=False,
            aggregations=aggregations,
            sort=sort or "_score:desc,id:asc",
            track_total_hits=True,
        )

    async def _index_exists_request(self, index_name: str) -> bool:
        return await self.client.indices.exists(index=index_name)

    async def _bulk_op_request(self, actions: List[Dict[str, Any]]):
        # https://www.elastic.co/guide/en/elasticsearch/reference/current/indices-refresh.html#refresh-api-desc
        _, errors = await helpers.async_bulk(client=self.client, actions=actions, raise_on_error=False, refresh=True)
        if errors:
            raise RuntimeError(errors)

    async def _refresh_index_request(self, index_name: str):
        await self.client.indices.refresh(index=index_name)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import AsyncGenerator

import pytest
import pytest_asyncio
from argilla_server.search_engine import ElasticSearchEngine, OpenSearchEngine
from argilla_server.settings import settings


@pytest.fixture
def search_engine(request):
    if settings.search_engine == "elasticsearch":
        engine = "elasticsearch_engine"
    elif settings.search_engine == "opensearch":
        engine = "opensearch_engine"
    else:
        raise Exception(f"Unknown search engine: {settings.search_engine}")

    return request.getfixturevalue(engine)


@pytest_asyncio.fixture()
async def elasticsearch_engine(elasticsearch_config: dict) -> AsyncGenerator[ElasticSearchEngine, None]:
    engine = ElasticSearchEngine(config=elasticsearch_config, number_of_replicas=0, number_of_shards=1)
    yield engine

    await engine.client.close()


@pytest_asyncio.fixture()
async def opensearch_engine(elasticsearch_config: dict) -> AsyncGenerator[OpenSearchEngine, None]:
    engine = OpenSearchEngine(config=elasticsearch_config, number_of_replicas=0, number_of_shards=1)
    yield engine

    await engine.client.close()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, List, TypeVar

from argilla_server.daos.backend.search.model import BaseRecordsQuery, QueryRange, SortableField, SortConfig
from argilla_server.pydantic_v1 import BaseModel, Field
from argilla_server.services.tasks.commons import ServiceRecord


class ServiceBaseRecordsQuery(BaseRecordsQuery):
    pass


class ServiceSortConfig(SortConfig):
    pass


class ServiceSortableField(SortableField):
    """Sortable field structure"""

    pass


class ServiceQueryRange(QueryRange):
    pass


class ServiceScoreRange(ServiceQueryRange):
    pass


class ServiceBaseSearchResultsAggregations(BaseModel):
    predicted_as: Dict[str, int] = Field(default_factory=dict)
    annotated_as: Dict[str, int] = Field(default_factory=dict)
    annotated_by: Dict[str, int] = Field(default_factory=dict)
    predicted_by: Dict[str, int] = Field(default_factory=dict)
    status: Dict[str, int] = Field(default_factory=dict)
    predicted: Dict[str, int] = Field(default_factory=dict)
    score: Dict[str, int] = Field(default_factory=dict)
    words: Dict[str, int] = Field(default_factory=dict)
    metadata: Dict[str, Dict[str, Any]] = Field(default_factory=dict)


ServiceSearchResultsAggregations = TypeVar(
    "ServiceSearchResultsAggregations", bound=ServiceBaseSearchResultsAggregations
)


class ServiceSearchResults(BaseModel):
    total: int
    records: List[ServiceRecord]
    metrics: Dict[str, Any] = Field(default_factory=dict)


ServiceRecordsQuery = TypeVar("ServiceRecordsQuery", bound=ServiceBaseRecordsQuery)
